from ctapipe.utils import get_dataset
from ctapipe.io.hessio import hessio_event_source
from ctapipe.core import Component
from traitlets import Unicode


class SimTelArrayReader(Component):

    """`SimTelArrayReader` class represents a Producer for pipeline.
        It opens simtelarray file and yields even to the next stage
    """
    filename = Unicode('gamma_test.simtel.gz', help='simtel MC input file').tag(
        config=True, allow_none=True)
    source = None

    def init(self):
        self.log.debug("--- SimTelArrayReader init {}---".format(self.filename))
        try:
            in_file = get_dataset(self.filename)
            self.source = hessio_event_source(in_file,max_events=3)
            self.log.debug('{} successfully opened {}'.format(self.filename,self.source))
        except:
            self.log.error('could not open ' + in_file)
            return False
        return True

    def run(self):
        for event in self.source:
            self.log.debug('\n--- SimTelArrayReader send event {}'.format(event.dl0.event_id))
            yield (event)
        self.log.debug("\n--- SimTelArrayReader Done ---")

    def finish(self):
        self.log.debug("--- SimTelArrayReader finish ---")
        pass



from .camera import CameraGeometry
from .atmosphere import get_atmosphere_profile_table, get_atmosphere_profile_functions
from .telescope import TelescopeDescription
from .optics import OpticsDescription
from .subarray import SubarrayDescription


__all__ = ['CameraGeometry',
           'get_atmosphere_profile_functions', 'TelescopeDescription',
           'OpticsDescription','SubarrayDescription']


from ..linalg import rotation_matrix_2d
from numpy import dot, allclose, identity


def test_rotation_matrix_2d():

    # test that 360 rotation is back to the identity:
    assert allclose(rotation_matrix_2d('360d'), identity(2))

    # test that a vector can be rotated correcly:
    vec = [1, 0]
    mat = rotation_matrix_2d('90d')
    vecp = dot(vec, mat)
    assert allclose(vecp, [0, -1]), 'vector rotation is wrong'

    # test that the rotation is Hermitian

    m = rotation_matrix_2d('25d')
    assert allclose(dot(m, m.T),  identity(2)), "rotation should be Hermetian"



from astropy.io import fits
from traitlets.config.loader import Config
from json import load
import logging
import warnings

logger = logging.getLogger(__name__)

__all__ = ['traitlets_config_to_fits', 'json_to_fits']


def traitlets_config_to_fits(config, fits_filename, clobber=True):
    '''Write a FITS file that represents configuration.

    Parameters
    ----------
    config : traitlets.config.loader.Config
        a traitlets.config.loader.Config to write in FITS format
    fits_filename : str
        FITS file name to write
    clobber : bool
        When True, overwrite the output file if exists.

    Raises
    ------
    OSError : If FITS file containing the traitlets config is not written
    '''

    if not isinstance(config, Config):
        raise TypeError('Config must be an instance of traitlets.config.loader.Config')

    # hduList will contain one TableHDU per section
    hduList = fits.HDUList()
    # get all Configuration entries
    # loop over section
    for section, entry in config.items():
        header = fits.Header()
        for key, value in entry.items():

            # CONTINUE and HIERARCH are incompatible, so we have to decide
            # to either truncate key or value. I went for the key. @maxnoe
            if isinstance(value, str):
                if len(key) > 8 and (len(key) + len(value)) > 70:
                    warnings.warn(
                        'Key "{}" will be truncated to {}'.format(key, key[:8])
                    )
                    key = key[:8]

            header[key] = value

        table_0 = fits.TableHDU(data=None, header=header, name=section)
        hduList.append(table_0)
    try:
        hduList.writeto(fits_filename, clobber=clobber)
    except OSError as e:
        logging.exception('Could not do save {}'.format(fits_filename))
        raise


def json_to_fits(json_filename, fits_filename, clobber=True):
    '''Write a FITS file that represents json file for traitlets configuration.

    Parameters
    ----------
    json_filename : str
        json file name containing traitlets Configuration.
        Only one level of section is allowed.
    fits_filename : str
        FITS file name to write
    clobber : bool
        When True, overwrite the output file if exists.

    Raises
    ------
    OSError : if FITS file containing a copy of json content is not written
    FileNotFoundError : if fits_filename could not be open
    '''
    try:
        f = open(json_filename, 'r')
        # hduList will contain one TableHDU per section
        hduList = fits.HDUList()
        json_object = load(f)
        # Create a global header for key/value that not depend of a section
        global_header = fits.Header()

        # loop over json entries (corresponding to Python class or general purpose)
        for section, entry in json_object.items():
            header = fits.Header()

            if isinstance(entry, dict):
                for key, value in entry.items():
                    if isinstance(value, dict):
                        raise ValueError(
                            'Malformed json file: fits header cannot contain subobjects'
                        )

                    # CONTINUE and HIERARCH are incompatible, so we have to
                    # decide to either truncate key or value.
                    # I went for the key. @maxnoe
                    if isinstance(value, str):
                        if len(key) > 8 and (len(key) + len(value)) > 70:
                            warnings.warn(
                                'Key "{}" will be truncated to {}'.format(key, key[:8])
                            )
                            key = key[:8]

                    header[key] = value

                # create a new TableHDU with current header and append it to hduList
                table_0 = fits.TableHDU(data=None, header=header, name=section)
                hduList.append(table_0)

            else:
                # entry is not a dictionary of key value but a simple value.
                # That means last section is not a section but a general purpose entry
                global_header[section] = entry

        # create a new TableHDU for global_header and append it to hduList
        table_0 = fits.TableHDU(data=None, header=global_header, name='GLOBAL')
        hduList.append(table_0)
        # write hduList to FITS file
        try:
            hduList.writeto(fits_filename, clobber=clobber)
        except OSError as e:
            logging.exception('Could not do save {}'.format(fits_filename))
            raise

    except FileNotFoundError:
        logging.exception('Could not open  {}'.format(fits_filename))
        raise



"""
Classes pertaining to the description of a Cherenkov Telescope

Todo:
-----

- add more info in OpticsDescription (mirror area, facets, etc). How to guess
  this?
- add ability to write to/from tables (like that written by
  ctapipe-dump-instrument)
- add ability to construct by names TelescopeDescription.from_name(
  camera='LSTCam', optics=('SST','1M')) (which would create a very unbalanced
  telescope :-))

"""

from .optics import OpticsDescription
from .camera import CameraGeometry


class TelescopeDescription:
    """
    Describes a Cherenkov Telescope and it's associated `OpticsDescription` and
    `CameraGeometry`

    The string representation is a combination of the optics and
    camera, separated by a colon: "optics:camera" (e.g. "SST-1m:DigiCam")

    The `TelescopeDescription.guess()` constructor can be used to fill in
    info from metadata, e.g. for Monte-Carlo files.

    Parameters
    ----------
    optics: OpticsDescription
       the optics associated with this telescope
    camera: CameraGeometry
       the camera associated with this telescope
    """


    def __init__(self,
                 optics: OpticsDescription,
                 camera: CameraGeometry):

        self._optics = optics
        self._camera = camera

    @property
    def optics(self):
        """ OpticsDescription for this telescope """
        return self._optics

    @property
    def camera(self):
        """ CameraGeometry for this telescope"""
        return self._camera

    @classmethod
    def guess(cls, pix_x, pix_y, effective_focal_length):
        """
        Construct a TelescopeDescription from metadata, filling in the
        missing information using a lookup table.

        Parameters
        ----------
        pix_x: array
           array of pixel x-positions with units
        pix_y: array
           array of pixel y-positions with units
        effective_focal_length: float
           effective focal length of telescope with units (m)
        """
        camera = CameraGeometry.guess(pix_x, pix_y, effective_focal_length)
        optics = OpticsDescription.guess(effective_focal_length)
        return cls(optics=optics, camera=camera)


    def __str__(self):
        return str(self.optics) + ":" + str(self.camera)

    def __repr__(self):
        return "{}(optics={}, camera={})".format(self.__class__.__name__,
                                                 str(self.optics),
                                                 str(self.camera))


# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Qt QApplication and QMainWindow for GUI
This requires the pyside python library to be installed
"""
from ctapipe.flow.gui.graphwidget import GraphWidget
from ctapipe.flow.gui.infolabel import InfoLabel
from ctapipe.flow.gui.guiconnection import GuiConnexion
import ctapipe.flow.gui.images_rc
from PyQt4.QtGui import QMainWindow
from PyQt4.QtGui import QPushButton
from PyQt4.QtGui import QApplication
from PyQt4.QtGui import QPalette
from PyQt4.QtGui import QPixmap
from PyQt4.QtGui import QWidget
from PyQt4.QtGui import QColor
from PyQt4.QtGui import QGridLayout
from PyQt4.QtGui import QMenuBar
from PyQt4.QtGui import QMenu
from PyQt4.QtGui import QStatusBar
from PyQt4.QtGui import QAction
from PyQt4.QtGui import QLabel
from PyQt4.QtCore import Qt
from PyQt4.QtCore import QRect
from PyQt4.QtCore import QObject
from PyQt4.QtCore import QMetaObject
from PyQt4.QtCore import SIGNAL

class MainWindow(QMainWindow, object):
    """
    QMainWindow displays pipeline
    Parameters
    ----------
    port : str
        used port to communicate with pipeline
        Note: The firewall must be configure to accept input/output on this port
    """

    def __init__(self, port):
        super(MainWindow, self).__init__()
        self.setupUi(port)

    def setupUi(self, port):
        self.setObjectName("MainWindow")
        self.resize(600,600)
        self.centralwidget = QWidget(self)
        p = self.centralwidget.palette()
        self.centralwidget.setAutoFillBackground(True)
        p.setColor(self.centralwidget.backgroundRole(), QColor(126, 135, 152))
        self.centralwidget.setPalette(p)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(self)
        self.menubar.setGeometry(QRect(0, 0, 808, 25))
        self.menubar.setObjectName("menubar")
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(self)
        self.statusbar.setObjectName("statusbar")
        self.setStatusBar(self.statusbar)
        self.actionQuit = QAction(self)
        self.actionQuit.setObjectName("actionQuit")
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        self.menubar.addAction(self.menuFile.menuAction())
        self.actionReset = QAction(self)
        self.actionReset.setObjectName("reset")
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionReset)
        self.menubar.addAction(self.menuFile.menuAction())
        # add other GUI objects
        self.graph_widget = GraphWidget(self.statusbar)
        self.gridLayout.addWidget(self.graph_widget, 1, 11, 10, 10 )
        pixmap = QPixmap(':/images/cta-logo-mini.png')
        lbl = QLabel()
        lbl.setPixmap(pixmap)
        self.gridLayout.addWidget(lbl, 0, 0)
        p = self.graph_widget.palette()
        self.graph_widget.setAutoFillBackground(True)
        p.setColor(
            self.graph_widget.backgroundRole(),QColor(255,255,255))# QColor(226, 235, 252))
        self.graph_widget.setPalette(p)
        self.quitButton = QPushButton()  # self.centralwidget)
        self.quitButton.setObjectName("quitButton")
        self.quitButton.setText(QApplication.translate
                                ("MainWindow", "Quit", None, QApplication.UnicodeUTF8))
        self.gridLayout.addWidget(self.quitButton, 12, 0, 1, 1)
        self.info_label = InfoLabel(0,4)
        self.info_label.setAutoFillBackground(True)
        self.gridLayout.addWidget(self.info_label,1, 0, 1, 5)
        #self.info_label.setAlignment(PyQt4.Qt.AlignCenter);
        palette = QPalette()
        palette.setColor(self.info_label.backgroundRole(),Qt.lightGray)
        self.info_label.setPalette(palette)
        QObject.connect(
            self.quitButton, SIGNAL("clicked()"), self.stop)
        QObject.connect(
            self.actionQuit, SIGNAL("triggered()"), self.stop)
        QMetaObject.connectSlotsByName(self)
        self.retranslateUi()
        QObject.connect(
            self.actionQuit, SIGNAL("triggered()"), self.close)
        QMetaObject.connectSlotsByName(self)
        # Create GuiConnexion for ZMQ comminucation with pipeline
        self.guiconnection = GuiConnexion(gui_port=port, statusBar=self.statusbar)
        self.guiconnection.message.connect(self.graph_widget.pipechange)
        self.guiconnection.message.connect(self.info_label.pipechange)
        self.guiconnection.reset_message.connect(self.graph_widget.reset)
        self.guiconnection.reset_message.connect(self.info_label.reset)
        self.guiconnection.mode_message.connect(self.info_label.mode_receive)
        QObject.connect(
            self.actionReset, SIGNAL("triggered()"), self.guiconnection.reset)
        QMetaObject.connectSlotsByName(self)
        # start the process
        self.guiconnection.start()

    def retranslateUi(self):
        self.setWindowTitle(QApplication.translate(
            "ctapipe flow based GUI", "ctapipe flow based GUI", None, QApplication.UnicodeUTF8))
        self.menuFile.setTitle(QApplication.translate(
            "MainWindow", "File", None, QApplication.UnicodeUTF8))
        self.actionQuit.setText(QApplication.translate(
            "MainWindow", "Quit", None, QApplication.UnicodeUTF8))
        self.actionReset.setText(QApplication.translate(
            "MainWindow", "Reset", None, QApplication.UnicodeUTF8))

    def stop(self):
        """Method connect (via Qt slot) to exit button
        Stops self.guiconnection (for ZMQ communication) process.
        Close main_windows
        """
        self.guiconnection.finish()
        self.guiconnection.join()
        self.close()

    def closeEvent(self, event):
            self.guiconnection.finish()
            self.guiconnection.join()
            event.accept()  # let the window close

class ModuleApplication(QApplication):
    """
    QApplication
    Parameters
    ----------
    QApplication : QApplication
    """
    def __init__(self,  argv, port):
        super(ModuleApplication, self).__init__(argv)
        self.main_windows = MainWindow(port)
        self.main_windows.show()
        self.exec_()



from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset
from ctapipe.plotting.camera import CameraPlotter
import numpy as np


def test_eventplotter():
    dataset = get_dataset("gamma_test.simtel.gz")
    source = hessio_event_source(dataset, max_events=1)
    event = next(source)
    del source

    telid = list(event.r0.tels_with_data)[0]

    data = event.r0.tel[telid].adc_samples[0]
    plotter = CameraPlotter(event)

    camera = plotter.draw_camera(telid, data[:, 0])
    assert camera is not None
    np.testing.assert_array_equal(camera.image, data[:, 0])

    plotter.draw_camera_pixel_ids(telid, [0, 1, 2])

    waveform = plotter.draw_waveform(data[0, :])
    assert waveform is not None
    np.testing.assert_array_equal(waveform.get_ydata(), data[0, :])

    line = plotter.draw_waveform_positionline(0)
    assert line is not None
    np.testing.assert_array_equal(line.get_xdata(), [0, 0])


if __name__ == '__main__':
    test_eventplotter()


"""
Table interpolation class, to allow interpolation of 2D images in any number of other 
dimensions. Reads in an interim FITS table as defined below:

Tables organised as a standard FITS image file with each table as an image HDU. 
---------------
First (primary) HDU must contain the following header entries:

CRPIXx, CRVALx, CRDELTAx: Number, position and pixel spacing of the reference pixel in 
the image. Where x is the axis number (1 or 2).

GRIDVALS: Names of values describing the grid points
e.g. GRIDVALS = ALT,AZ
---------------
Each HDU must contain a header entry containing the grid points described by GRIDVALS
e.g. ALT =70
AZI = 0

It is also strongly recommended to include documentation for each interpolation 
dimension containing at least the units. This entry ust begin with DOC, followed by the 
dimension name
e.g. DOCALT = "Altitude of event (deg)"

TODO:
    - Improve error handling
    - Add option for caching nd interpolated value
    - Better deal with edges of phase space
    - Allow non-linear interpolation
"""

from scipy import interpolate
import numpy as np
from astropy.io import fits


class TableInterpolator:
    """
    This is a simple class for loading lookup tables from a fits file and
    interpolating between them
    """

    def __init__(self, filename, verbose=1):
        """
        Initialisation of class to load templates from a file and create the interpolation
        objects

        Parameters
        ----------
        filename: string
            Location of Template file
        verbose: int
            Verbosity level,
            0 = no logging
            1 = File + interpolation point information
            2 = Detailed description of interpolation points
        """
        self.verbose = verbose
        if self.verbose:
            print("Loading lookup tables from", filename)

        grid, bins, template = self.parse_fits_table(filename)
        x_bins, y_bins = bins

        self.interpolator = interpolate.LinearNDInterpolator(grid, template, fill_value=0)
        self.nearest_interpolator = interpolate.NearestNDInterpolator(grid, template)

        self.grid_interp = interpolate.RegularGridInterpolator((x_bins, y_bins),
                                                               np.zeros([x_bins.shape[0],
                                                                         y_bins.shape[
                                                                             0]]),
                                                               method="linear",
                                                               bounds_error=False,
                                                               fill_value=0)

    def parse_fits_table(self, filename):
        """
        Function opens tables contained within fits files and parses them into a format 
        recognisable by the interpolator.
        
        Parameters
        ----------
        filename: str
            Name of table file

        Returns
        -------
            tuple (grid points, bin centres, images)
        """
        file = fits.open(filename)
        template = list()
        grid = list()

        primHDU = file[0].header  # We require first HDU to be primary

        # Below definitions are standard
        ix, iy = primHDU["CRPIX2"], primHDU["CRPIX1"]
        val_x, val_y = primHDU["CRVAL2"], primHDU["CRVAL1"]
        print(val_x, val_y)
        delta_x, delta_y = primHDU["CRDELTA2"], primHDU["CRDELTA1"]
        nbins_x, nbins_y = primHDU["NAXIS2"], primHDU["NAXIS1"]
        ix *= delta_x
        iy *= delta_y

        x_bins = np.arange(val_x-ix ,val_x+(delta_x*nbins_x)-ix, step=delta_x)
        y_bins = np.arange(val_y-iy ,val_y+(delta_y*nbins_y)-iy, step=delta_y)
        grid_vals = primHDU["GRIDVALS"]
        points = grid_vals.split(",")

        if self.verbose:
            print("Interpolation point source be called in order", points)
        if self.verbose>1:
            for p in points:
                print(p,":", primHDU["DOC"+p])

        for hdu in file:
            template.append(hdu.data)
            
            hdu_pt = list()
            for p in points:
                hdu_pt.append(hdu.header[p])
            grid.append(np.array(hdu_pt))

        print(np.array(grid))
        bins = (x_bins, y_bins)
        return grid, bins, template

    def interpolate(self, params, pixel_pos_x, pixel_pos_y):
        """

        Parameters
        ----------
        params: ndarray
            numpy array of interpolation parameters
            currently [energy, impact distance, xmax]
        pixel_pos_x: ndarray
            pixel position in degrees
        pixel_pos_y: ndarray
            pixel position in degrees

        Returns
        -------
        ndarray of expected intensity for all pixel positions given

        """

        image = self.interpolated_image(params)
        self.grid_interp.values = image

        points = np.array([pixel_pos_x, pixel_pos_y])
        return self.grid_interp(points.T)

    def interpolated_image(self, params):
        """
        Function for creating a ful interpolated image template from the interpolation library

        Parameters
        ----------
        params: ndarray
            numpy array of interpolation parameters
            currently [energy, impact distance, xmax]

        Returns
        -------
        ndarray of a single image template

        """
        image = self.interpolator(params)[0]
        if np.isnan(image).all():
            print("Found a NaN", params)
            image = self.nearest_interpolator(params)[0]

        return image



from ctapipe.core import Component
from time import sleep

class Add(Component):
    """Add class represents a Stage for pipeline.
       It simply adds one to the received value and returned it.

    """
    def init(self):
        self.log.debug("--- Add init ---")
        return True

    def run(self, x):
        sleep(.5)
        self.log.debug("Add receive {} ".format(x))
        return x

    def finish(self):
        self.log.debug("--- Add finish ---")
        pass



from ctapipe.reco.shower_max import ShowerMaxEstimator
from astropy import units as u
from ctapipe.utils import get_dataset


def test_showermaxestimator(en=5*u.TeV, h_first_int=10*u.km, az=70*u.deg):
    estim = ShowerMaxEstimator(atmosphere_profile_name='paranal')
    h_max = estim.find_shower_max_height(en, h_first_int, az)
    assert h_max.unit.is_equivalent(u.m), "return value has not proper unit"
    return h_max

if __name__ == "__main__":
    en, h_first_int, az = 5*u.TeV, 10*u.km, 70*u.deg
    print("ShowerMaxEstimator unit test:")
    print("Energy = {}, hight of first interaction = {}, azimuth = {}"
          .format(en, h_first_int, az))
    print("h_max:", test_showermaxestimator())



import numpy as np

from astropy import units as u

from .regressor_classifier_base import RegressorClassifierBase
from sklearn.ensemble import RandomForestRegressor


class EnergyRegressor(RegressorClassifierBase):
    """This class collects one regressor for every camera type -- given
    by `cam_id_list` -- to get an estimate for the energy of an
    air-shower.  The class interfaces with `scikit-learn` in that it
    relays all function calls not exeplicitly defined here through
    `.__getattr__` to any regressor in its collection.  This gives us
    the full power and convenience of scikit-learn and we only have to
    bother about the high-level interface.

    The fitting approach is to train on single images; then separately
    predict an image's energy and combine the different predictions
    for all the telscopes in each event in one single energy estimate.

    TODO: come up with a better merging approach to combine the
          estimators of the various camera types

    Parameters
    ----------
    regressor : scikit-learn regressor
        the regressor you want to use to estimate the shower energies
    cam_id_list : list of strings
        list of identifiers to differentiate the various sources of
        the images; could be the camera IDs or even telescope IDs. We
        will train one regressor for each of the identifiers.
    unit : astropy.Quantity
        scikit-learn regressors don't work with astropy unit. so, tell
        in advance in which unit we want to deal here.
    kwargs
        arguments to be passed on to the constructor of the regressors

    """

    def __init__(self, regressor=RandomForestRegressor,
                 cam_id_list=("cam"), unit=u.TeV, **kwargs):
        super().__init__(model=regressor, cam_id_list=cam_id_list,
                         unit=unit, **kwargs)

    def predict_by_event(self, X):
        """expects a list of events where every "event" is a dictionary
        mapping telescope identifiers to the list of feature-lists by
        the telescopes of that type in this event. Refer to the Note
        section in `.reshuffle_event_list` for an description how `X`
        is supposed to look like.  The singular estimate for the event
        is simply the mean of the various estimators of the event.

        X : list of "events"
            cf. `.reshuffle_event_list` under Notes

        Returns
        -------
        dict :
            dictionary that contains various statistical modes (mean,
            median, standard deviation) of the predicted quantity of
            every telescope for all events.

        Raises
        ------
        KeyError:
            if there is a telescope identifier in `X` that is not a
            key in the regressor dictionary

        """

        predict_mean = []
        predict_median = []
        predict_std = []
        for evt in X:
            predicts = []
            weights = []
            for cam_id, tels in evt.items():
                try:
                    t_res = self.model_dict[cam_id].predict(tels).tolist()
                    predicts += t_res
                except KeyError:
                    # QUESTION if there is no trained classifier for
                    # `cam_id`, raise an error or just pass this
                    # camera type?
                    raise KeyError("cam_id '{}' in X but no model defined: {}"
                                   .format(cam_id, [k for k in self.model_dict]))

                try:
                    # if a `namedtuple` is provided, we can weight the different images
                    # using some of the provided features
                    weights += [t.sum_signal_cam / t.impact_dist for t in tels]
                except AttributeError:
                    # otherwise give every image the same weight
                    weights += [1] * len(tels)

            predict_mean.append(np.average(predicts, weights=weights))
            predict_median.append(np.median(predicts))
            predict_std.append(np.std(predicts))

        return {"mean": np.array(predict_mean) * self.unit,
                "median": np.array(predict_median) * self.unit,
                "std": np.array(predict_std) * self.unit}

    def predict_by_telescope_type(self, X):
        """same as `predict_dict` only that it returns a list of dictionaries
        with an estimate for the target quantity for every telescope
        type separately.

        more for testing- and performance-measuring-purpouses -- to
        see how the different telescope types behave throughout the
        energy ranges and if a better (energy-dependant) combination
        of the separate telescope-wise estimators (compared to the
        mean) can be achieved.

        """

        predict_list_dict = []
        for evt in X:
            res_dict = {}
            for cam_id, tels in evt.items():
                t_res = self.model_dict[cam_id].predict(tels).tolist()
                res_dict[cam_id] = np.mean(t_res) * self.unit
            predict_list_dict.append(res_dict)

        return predict_list_dict

    @classmethod
    def load(cls, path, cam_id_list, unit=u.TeV):
        """this is only here to overwrite the unit argument with an astropy
        quantity

        Parameters
        ----------
        path : string
            the path where the pre-trained, pickled regressors are
            stored `path` is assumed to contain a `{cam_id}` keyword
            to be replaced by each camera identifier in `cam_id_list`
            (or at least a naked `{}`).
        cam_id_list : list
            list of camera identifiers like telescope ID or camera ID
            and the assumed distinguishing feature in the filenames of
            the various pickled regressors.
        unit : astropy.Quantity
            scikit-learn regressor do not work with units. so append
            this one to the predictions. assuming that the models
            where trained with consistent units. (default: u.TeV)

        Returns
        -------
        EnergyRegressor:
            a ready-to-use instance of this class to predict any
            quantity you have trained for

        """
        return super().load(path, cam_id_list, unit)



import numpy as np
from ctapipe.image import cleaning

from ctapipe.instrument import CameraGeometry


def test_tailcuts_clean():

    geom = CameraGeometry.from_name("LSTCam")
    image = np.zeros_like(geom.pix_id, dtype=np.float)
    pedvar = np.ones_like(geom.pix_id, dtype=np.float)

    N = 40
    some_neighs = geom.neighbors[N][0:3]  # pick 4 neighbors
    image[N] = 5.0              # set a single image pixel
    image[some_neighs] = 3.0    # make some boundaries that are neighbors
    image[10] = 3.0             # a boundary that is not a neighbor

    mask = cleaning.tailcuts_clean(geom, image, picture_thresh=4.5,
                                   boundary_thresh=2.5)

    print((mask > 0).sum(), "clean pixels")
    print(geom.pix_id[mask])

    assert 10 not in geom.pix_id[mask]
    assert set(some_neighs).union({N}) == set(geom.pix_id[mask])
    assert (mask > 0).sum() == 4


def test_dilate():

    geom = CameraGeometry.from_name("LSTCam")
    mask = np.zeros_like(geom.pix_id, dtype=bool)

    mask[100] = True  # a single pixel far from a border is true.
    assert mask.sum() == 1

    # dilate a single row
    dmask = cleaning.dilate(geom, mask)
    assert dmask.sum() == 1 + 6

    # dilate a second row
    dmask = cleaning.dilate(geom, dmask)
    assert dmask.sum() == 1 + 6 + 12

    # dilate a third row
    dmask = cleaning.dilate(geom, dmask)
    assert dmask.sum() == 1 + 6 + 12 + 18

def test_tailcuts_clean():

    # start with simple 3-pixel camera
    geom = CameraGeometry.make_rectangular(3, 1, (-1, 1))

    p = 15  #picture value
    b = 7   # boundary value

    # for no isolated pixels:
    testcases = {(p, p, 0): [True, True, False],
                 (p, 0, p): [False, False, False],
                 (p, b, p): [True, True, True],
                 (p, b, 0): [True, True, False],
                 (b, b, 0): [False, False, False],
                 (0, p ,0): [False, False, False]}

    for image, mask in testcases.items():
        result = cleaning.tailcuts_clean(geom, np.array(image),
                                         picture_thresh=15,
                                         boundary_thresh=5,
                                         keep_isolated_pixels=False)
        assert (result == mask).all()

    # allowing isolated pixels
    testcases = {(p, p, 0): [True, True, False],
                 (p, 0, p): [True, False, True],
                 (p, b, p): [True, True, True],
                 (p, b, 0): [True, True, False],
                 (b, b, 0): [False, False, False],
                 (0, p, 0): [False, True, False]}

    for image, mask in testcases.items():
        result = cleaning.tailcuts_clean(geom, np.array(image),
                                         picture_thresh=15,
                                         boundary_thresh=5,
                                         keep_isolated_pixels=True)
        assert (result == mask).all()



from ctapipe.core import Component
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from traitlets import Bool
from traitlets import Unicode

class DisplayTelescopeStep(Component):
    """DisplayTelescopeStep` class represents a Stage for pipeline.
        it displays calibrated camera events with matplotlib and/or write
        them to a PDF file
    """
    display = Bool(default_value=False, help='Display the camera events') \
     .tag(config=True)
    pdf_path = Unicode(default_value=None,allow_none=True,
        help='Path to store a pdf output of the plots').tag(config=True)

    def init(self):
        self.log.debug("--- DisplayTelescopeStep init ---")
        self.pdfPages = None
        if self.pdf_path:
            self.pdfPages = PdfPages(self.pdf_path)
        return True

    def run(self,figure):
            self.log.debug("DisplayTelescopeStep receive {}".format(figure))
            if self.display:
                plt.pause(.1)
            if self.pdfPages is not None:
                self.pdfPages.savefig(figure)

    def finish(self):
        self.log.debug("--- DisplayTelescopeStep finish ---")
        if self.pdfPages:
            self.pdfPages.close()
        pass



"""
Container structures for data that should be read or written to disk
"""

from astropy import units as u
from astropy.time import Time
from numpy import ndarray

from ..core import Container, Field, Map
from ..instrument import SubarrayDescription

__all__ = ['InstrumentContainer',
           'R0Container',
           'R0CameraContainer',
           'R1Container',
           'R1CameraContainer',
           'DL0Container',
           'DL0CameraContainer',
           'DL1Container',
           'DL1CameraContainer',
           'MCEventContainer',
           'MCHeaderContainer',
           'MCCameraEventContainer',
           'CameraCalibrationContainer',
           'CentralTriggerContainer',
           'ReconstructedContainer',
           'ReconstructedShowerContainer',
           'ReconstructedEnergyContainer',
           'ParticleClassificationContainer',
           'DataContainer',
           'HillasParametersContainer']

# todo: change some of these Maps to be just 3D NDarrays?


class InstrumentContainer(Container):
    """Storage of header info that does not change with event. This is a
    temporary hack until the Instrument module and database is fully
    implemented.  Eventually static information like this will not be
    part of the data stream, but be loaded and accessed from
    functions.

    """

    subarray = Field(SubarrayDescription("MonteCarloArray"),
                     "SubarrayDescription from the instrument module")

    telescope_ids = Field([], "list of IDs of telescopes used in the run")
    pixel_pos = Field(Map(ndarray), "map of tel_id to pixel positions")
    optical_foclen = Field(Map(ndarray), "map of tel_id to focal length")
    mirror_dish_area = Field(Map(float), "map of tel_id to the area of the mirror dish", unit=u.m**2)
    mirror_numtiles = Field(Map(int), "map of tel_id to the number of tiles for the mirror")
    tel_pos = Field(Map(ndarray), "map of tel_id to telescope position")
    num_pixels = Field(Map(int), "map of tel_id to number of pixels in camera")
    num_channels = Field(Map(int), "map of tel_id to number of channels")




class DL1CameraContainer(Container):
    """Storage of output of camera calibration e.g the final calibrated
    image in intensity units and other per-event calculated
    calibration information.
    """
    image = Field(None, "np array of camera image", unit=u.electron)
    extracted_samples = Field(None, (
        "numpy array of bools indicating which samples were included in the "
        "charge extraction as a result of the charge extractor chosen. "
        "Shape=(nchan, npix, nsamples)."
    ))
    peakpos = Field(None, (
        "numpy array containing position of the peak as determined by the "
        "peak-finding algorithm for each pixel and channel"
    ))
    cleaned = Field(
        None, "numpy array containing the waveform after cleaning"
    )


class CameraCalibrationContainer(Container):
    """
    Storage of externally calculated calibration parameters (not per-event)
    """
    dc_to_pe = Field(None, "DC/PE calibration arrays from MC file")
    pedestal = Field(None, "pedestal calibration arrays from MC file")


class DL1Container(Container):
    """ DL1 Calibrated Camera Images and associated data"""
    tel = Field(Map(DL1CameraContainer), "map of tel_id to DL1CameraContainer")


class R0CameraContainer(Container):
    """
    Storage of raw data from a single telescope
    """
    adc_sums = Field(None, (
        "numpy array containing integrated ADC data "
        "(n_channels x n_pixels)"
    ))
    adc_samples = Field(None, (
        "numpy array containing ADC samples"
        "(n_channels x n_pixels, n_samples)"
    ))
    num_samples = Field(None, "number of time samples for telescope")



class R0Container(Container):
    """
    Storage of a Merged Raw Data Event
    """

    run_id = Field(-1, "run id number")
    event_id = Field(-1, "event id number")
    tels_with_data = Field([], "list of telescopes with data")
    tel = Field(Map(R0CameraContainer), "map of tel_id to R0CameraContainer")


class R1CameraContainer(Container):
    """
    Storage of r1 calibrated data from a single telescope
    """
    pe_samples = Field(None, (
        "numpy array containing p.e. samples"
        "(n_channels x n_pixels, n_samples)"
    ))


class R1Container(Container):
    """
    Storage of a r1 calibrated Data Event
    """

    run_id = Field(-1, "run id number")
    event_id = Field(-1, "event id number")
    tels_with_data = Field([], "list of telescopes with data")
    tel = Field(Map(R1CameraContainer), "map of tel_id to R1CameraContainer")


class DL0CameraContainer(Container):
    """
    Storage of data volume reduced dl0 data from a single telescope
    """
    pe_samples = Field(None, (
        "numpy array containing data volume reduced "
        "p.e. samples"
        "(n_channels x n_pixels, n_samples)"
    ))


class DL0Container(Container):
    """
    Storage of a data volume reduced Event
    """

    run_id = Field(-1, "run id number")
    event_id = Field(-1, "event id number")
    tels_with_data = Field([], "list of telescopes with data")
    tel = Field(Map(DL0CameraContainer), "map of tel_id to DL0CameraContainer")


class MCCameraEventContainer(Container):
    """
    Storage of mc data for a single telescope that change per event
    """
    photo_electron_image = Field(
        Map(), "reference image in pure photoelectrons, with no noise"
    )
    # todo: move to instrument (doesn't change per event)
    reference_pulse_shape = Field(
        None, "reference pulse shape for each channel"
    )
    # todo: move to instrument or a static MC container (don't change per
    # event)
    time_slice = Field(0, "width of time slice", unit=u.ns)
    dc_to_pe = Field(None, "DC/PE calibration arrays from MC file")
    pedestal = Field(None, "pedestal calibration arrays from MC file")
    azimuth_raw = Field(
        0, "Raw azimuth angle [radians from N->E] for the telescope"
    )
    altitude_raw = Field(0, "Raw altitude angle [radians] for the telescope")
    azimuth_cor = Field(
        0, "the tracking Azimuth corrected for pointing errors for the telescope"
    )
    altitude_cor = Field(
        0, "the tracking Altitude corrected for pointing errors for the telescope"
    )


class MCEventContainer(Container):
    """
    Monte-Carlo
    """
    energy = Field(0.0, "Monte-Carlo Energy", unit=u.TeV)
    alt = Field(0.0, "Monte-carlo altitude", unit=u.deg)
    az = Field(0.0, "Monte-Carlo azimuth", unit=u.deg)
    core_x = Field(0.0, "MC core position", unit=u.m)
    core_y = Field(0.0, "MC core position", unit=u.m)
    h_first_int = Field(0.0, "Height of first interaction")
    tel = Field(
        Map(MCCameraEventContainer), "map of tel_id to MCCameraEventContainer"
    )


class MCHeaderContainer(Container):
    """
    Monte-Carlo information that doesn't change per event
    """
    run_array_direction = Field([], (
        "the tracking/pointing direction in "
        "[radians]. Depending on 'tracking_mode' "
        "this either contains: "
        "[0]=Azimuth, [1]=Altitude in mode 0, "
        "OR "
        "[0]=R.A., [1]=Declination in mode 1."
    ))


class CentralTriggerContainer(Container):

    gps_time = Field(Time, "central average time stamp")
    tels_with_trigger = Field([], "list of telescopes with data")


class ReconstructedShowerContainer(Container):
    """
    Standard output of algorithms reconstructing shower geometry
    """

    alt = Field(0.0, "reconstructed altitude", unit=u.deg)
    alt_uncert = Field(0.0, "reconstructed altitude uncertainty", unit=u.deg)
    az = Field(0.0, "reconstructed azimuth", unit=u.deg)
    az_uncert = Field(0.0, 'reconstructed azimuth uncertainty', unit=u.deg)
    core_x = Field(
        0.0, 'reconstructed x coordinate of the core position', unit=u.m
    )
    core_y = Field(
        0.0, 'reconstructed y coordinate of the core position', unit=u.m
    )
    core_uncert = Field(
        0.0, 'uncertainty of the reconstructed core position', unit=u.m
    )
    h_max = Field(0.0, 'reconstructed height of the shower maximum')
    h_max_uncert = Field(0.0, 'uncertainty of h_max')
    is_valid = Field(False, (
        'direction validity flag. True if the shower direction'
        'was properly reconstructed by the algorithm'
    ))
    tel_ids = Field([], (
        'list of the telescope ids used in the'
        ' reconstruction of the shower'
    ))
    average_size = Field(0.0, 'average size of used')
    goodness_of_fit = Field(0.0, 'measure of algorithm success (if fit)')


class ReconstructedEnergyContainer(Container):
    """
    Standard output of algorithms estimating energy
    """
    energy = Field(-1.0, 'reconstructed energy', unit=u.TeV)
    energy_uncert = Field(-1.0, 'reconstructed energy uncertainty', unit=u.TeV)
    is_valid = Field(False, (
        'energy reconstruction validity flag. True if '
        'the energy was properly reconstructed by the '
        'algorithm'
    ))
    tel_ids = Field([], (
        'array containing the telescope ids used in the'
        ' reconstruction of the shower'
    ))
    goodness_of_fit = Field(0.0, 'goodness of the algorithm fit')


class ParticleClassificationContainer(Container):
    """
    Standard output of gamma/hadron classification algorithms
    """
    # TODO: Do people agree on this? This is very MAGIC-like.
    # TODO: Perhaps an integer classification to support different classes?
    # TODO: include an error on the prediction?
    prediction = Field(0.0, (
        'prediction of the classifier, defined between '
        '[0,1], where values close to 0 are more '
        'gamma-like, and values close to 1 more '
        'hadron-like'
    ))
    is_valid = Field(False, (
        'classificator validity flag. True if the '
        'predition was successful within the algorithm '
        'validity range')
    )

    # TODO: KPK: is this different than the list in the reco
    # container? Why repeat?
    tel_ids = Field([], (
        'array containing the telescope ids used '
        'in the reconstruction of the shower'
    ))
    goodness_of_fit = Field(0.0, 'goodness of the algorithm fit')


class ReconstructedContainer(Container):
    """ collect reconstructed shower info from multiple algorithms """

    shower = Field(
        Map(ReconstructedShowerContainer),
        "Map of algorithm name to shower info"
    )
    energy = Field(
        Map(ReconstructedEnergyContainer),
        "Map of algorithm name to energy info"
    )
    classification = Field(
        Map(ParticleClassificationContainer),
        "Map of algorithm name to classification info"
    )


class DataContainer(Container):
    """ Top-level container for all event information """

    r0 = Field(R0Container(), "Raw Data")
    r1 = Field(R1Container(), "R1 Calibrated Data")
    dl0 = Field(DL0Container(), "DL0 Data Volume Reduced Data")
    dl1 = Field(DL1Container(), "DL1 Calibrated image")
    dl2 = Field(ReconstructedContainer(), "Reconstructed Shower Information")
    mc = Field(MCEventContainer(), "Monte-Carlo data")
    mcheader = Field(MCHeaderContainer(), "Monte-Carlo run header data")
    trig = Field(CentralTriggerContainer(), "central trigger information")
    count = Field(0, "number of events processed")
    inst = Field(InstrumentContainer(), "instrumental information (deprecated")


class MuonRingParameter(Container):
    """
    Storage of muon ring fit output

    Parameters
    ----------

    run_id : int
        run number
    event_id : int
        event number
    tel_id : int
        telescope ID
    ring_center_x, ring_center_y, ring_radius:
        center position and radius of the fitted ring
    ring_chi2_fit:
        chi squared of the ring fit
    ring_cov_matrix:
        covariance matrix of ring parameters
    """

    run_id = Field(0, "run identification number")
    event_id = Field(0, "event identification number")
    tel_id = Field(0, 'telescope identification number')
    ring_center_x = Field(0.0, 'centre (x) of the fitted muon ring')
    ring_center_y = Field(0.0, 'centre (y) of the fitted muon ring')
    ring_radius = Field(0.0, 'radius of the fitted muon ring')
    ring_chi2_fit = Field(0.0, 'chisquare of the muon ring fit')
    ring_cov_matrix = Field(0.0, 'covariance matrix of the muon ring fit')
    ring_fit_method = Field("", 'fitting method used for the muon ring')
    inputfile = Field("", 'input file')


class MuonIntensityParameter(Container):
    """
    Storage of muon intensity fit output

    Parameters
    ----------

    run_id : int
        run number
    event_id : int
        event number
    tel_id : int
        telescope ID
    impact_parameter: float
        reconstructed impact parameter
    impact_parameter_chi2:
        chi squared impact parameter
    intensity_cov_matrix:
        Covariance matrix of impact parameters or alternatively:
        full 5x5 covariance matrix for the complete fit (ring + impact)
    impact_parameter_pos_x, impact_parameter_pos_y:
        position on the mirror of the muon impact
    COG_x, COG_y:
        center of gravity
    optical_efficiency_muon:
        optical muon efficiency from intensity fit
    ring_completeness:
        completeness of the ring
    ring_num_pixel: int
        Number of pixels composing the ring
    ring_size:
        ring size
    off_ring_size:
        size outside of the ring
    ring_width:
        ring width
    ring_time_width:
        standard deviation of the photons time arrival

    prediction: dict
        ndarray of the predicted charge in all pixels
    mask:
        ndarray of the mask used on the image for fitting

    """
    run_id = Field(0, 'run identification number')
    event_id = Field(0, 'event identification number')
    tel_id = Field(0, 'telescope identification number')
    ring_completeness = Field(0., 'fraction of ring present')
    ring_num_pixel = Field(0, 'number of pixels in the ring image')
    ring_size = Field(0., 'size of the ring in pe')
    off_ring_size = Field(0., 'image size outside of ring in pe')
    ring_width = Field(0., 'width of the muon ring in degrees')
    ring_time_width = Field(0., 'duration of the ring image sequence')
    impact_parameter = Field(0., 'distance of muon impact position from centre of mirror')
    impact_parameter_chi2 = Field(0., 'impact parameter chi squared')
    intensity_cov_matrix = Field(0., 'covariance matrix of intensity')
    impact_parameter_pos_x = Field(0., 'impact parameter x position')
    impact_parameter_pos_y = Field(0., 'impact parameter y position')
    COG_x = Field(0.0, 'Centre of Gravity x')
    COG_y = Field(0.0, 'Centre of Gravity y')
    prediction = Field([], 'image prediction')
    mask = Field([], 'image pixel mask')
    optical_efficiency_muon = Field(0., 'optical efficiency muon')
    intensity_fit_method = Field("", 'intensity fit method')
    inputfile = Field("", 'input file')


class HillasParametersContainer(Container):

    intensity = Field(0.0, 'total intensity (size)')

    x = Field(0.0, 'centroid x coordinate')
    y = Field(0.0, 'centroid x coordinate')
    r = Field(0.0, 'radial coordinate of centroid')
    phi = Field(0.0, 'polar coordinate of centroid', unit=u.deg)

    length = Field(0.0, 'RMS spread along the major-axis')
    width = Field(0.0, 'RMS spread along the minor-axis')
    psi = Field(0.0, 'rotation angle of ellipse', unit=u.deg)

    skewness = Field(0.0, 'measure of the asymmetry')
    kurtosis = Field(0.0, 'measure of the tailedness')



import pytest
from traitlets import Float, TraitError

from ctapipe.core import Component
from abc import abstractmethod


def test_component_is_abstract():

    class AbstractComponent(Component):
        @abstractmethod
        def test(self):
            pass

    with pytest.raises(TypeError):
        AbstractComponent()


def test_component_simple():
    """
    very basic test to construct a component and check
    that it's traits work correctly
    """

    class ExampleComponent(Component):
        description = "this is a test"
        param = Float(default_value=1.0,
                      help="float parameter").tag(config=True)

    comp = ExampleComponent(None)

    assert comp.has_trait('param') is True
    comp.param = 1.2

    with pytest.raises(TraitError):
        comp.param = "badvalue"



# Licensed under a 3-clause BSD style license - see LICENSE.rst
""" print information about ctapipe and its command-line tools. """
import importlib
import logging
import os
import sys

import ctapipe_resources

from ..utils import datasets
from ..core import Provenance
from .utils import get_parser

__all__ = ['info']

# TODO: this list should be global (or generated at install time)
_dependencies = sorted(['astropy', 'matplotlib',
                        'numpy', 'traitlets',
                        'sklearn', 'scipy', 'numba',
                        'pytest', 'ctapipe_resources', 'iminuit', 'tables'])

_optional_dependencies = sorted(['pytest', 'graphviz', 'pyzmq',
                                 'fitsio', 'pyhessio', 'targetio',
                                 'matplotlib'])


def main(args=None):
    parser = get_parser(info)
    parser.add_argument('--version', action='store_true',
                        help='Print version number')
    parser.add_argument('--tools', action='store_true',
                        help='Print available command line tools')
    parser.add_argument('--dependencies', action='store_true',
                        help='Print available versions of dependencies')
    parser.add_argument('--resources', action='store_true',
                        help='Print available versions of dependencies')
    parser.add_argument('--system', action='store_true',
                        help='Print system info')
    parser.add_argument('--all', action='store_true',
                        help='show all info')
    args = parser.parse_args(args)

    if len(sys.argv) <= 1:
        parser.print_help()
        sys.exit(1)

    info(**vars(args))


def info(version=False, tools=False, dependencies=False,
         resources=False, system=False, all=False):
    """Print various info to the console.

    TODO: explain.
    """
    logging.basicConfig(level=logging.INFO,
                        format='%(levelname)s - %(message)s')

    if version or all:
        _info_version()

    if tools or all:
        _info_tools()

    if dependencies or all:
        _info_dependencies()

    if resources or all:
        _info_resources()

    if system or all:
        _info_system()


def _info_version():
    """Print version info."""
    import ctapipe
    print('\n*** ctapipe version info ***\n')
    print('version: {0}'.format(ctapipe.__version__))
    # print('release: {0}'.format(version.release))
    # print('githash: {0}'.format(version.githash))
    print('')


def _info_tools():
    """Print info about command line tools."""
    print('\n*** ctapipe tools ***\n')
    print('the following can be executed by typing ctapipe-<toolname>:')
    print('')

    # TODO: how to get a one-line description or
    # full help text from the docstring or ArgumentParser?
    # This is the function names, we want the command-line names
    # that are defined in setup.py !???
    from ctapipe.tools.utils import get_all_descriptions
    from textwrap import TextWrapper
    wrapper = TextWrapper(width=80,
                          subsequent_indent=" " * 35)

    scripts = get_all_descriptions()
    for name, desc in sorted(scripts.items()):
        text = "{:<30s}  - {}".format(name, desc)
        print(wrapper.fill(text))
        print('')
    print('')


def _info_dependencies():
    """Print info about dependencies."""
    print('\n*** ctapipe core dependencies ***\n')

    for name in _dependencies:
        try:
            module = importlib.import_module(name)
            version = module.__version__
        except ImportError:
            version = 'not installed'

        print('{:>20s} -- {}'.format(name, version))

    print('\n*** ctapipe optional dependencies ***\n')

    for name in _optional_dependencies:
        try:
            module = importlib.import_module(name)
            version = module.__version__
        except ImportError:
            version = 'not installed'
        except AttributeError:
            version = "installed, but __version__ doesn't exist"

        print('{:>20s} -- {}'.format(name, version))


def _info_resources():
    """ display all known resources """

    print('\n*** ctapipe resources ***\n')

    print("ctapipe_resources version: {}".format(ctapipe_resources.__version__))

    print("CTAPIPE_SVC_PATH: (directories where resources are searched)")
    if os.getenv('CTAPIPE_SVC_PATH') is not None:
        for dir in datasets.get_searchpath_dirs():
            print("\t * {}".format(dir))
    else:
        print("\t no path is set")
    print("")

    all_resources = sorted(datasets.find_all_matching_datasets("\w.*"))
    locations = [os.path.dirname(datasets.get_dataset(name))
                 for name in all_resources]
    home = os.path.expanduser("~")
    resource_dir = os.path.dirname(datasets.get_dataset(
        "HESS-I.camgeom.fits.gz"))

    fmt = "{name:<30.30s} : {loc:<30.30s}"
    print(fmt.format(name="RESOURCE NAME", loc="LOCATION"))
    print("-" * 70)
    for name, loc in zip(all_resources, locations):
        if name.endswith(".py") or name.startswith("_"):
            continue
        loc = loc.replace(resource_dir, "[ctapipe_resources]")
        loc = loc.replace(home, "~")
        print(fmt.format(name=name, loc=loc))


def _info_system():
    # collect system info using the ctapipe provenance system :

    print('\n*** ctapipe system environment ***\n')

    prov = Provenance()
    system_prov = prov.current_activity.provenance['system']

    for section in ['platform','python']:

        print('\n====== ',section," ======== \n")
        sysinfo = system_prov[section]

        for name, val in sysinfo.items():
            print("{:>20.20s} -- {:<60.60s}".format(name, str(val)))



from ctapipe.utils import datasets
import os
import pytest

def test_find_datasets():

    # find all datasets matching pattern
    r = datasets.find_all_matching_datasets("(.*)\.camgeom\.fits\.gz")
    assert len(r) > 3

    # get the full filename for a resrouces
    assert os.path.exists(datasets.get_dataset(r[0]))

    # try using a pattern
    r = datasets.find_all_matching_datasets("(.*)\.camgeom\.fits\.gz",
                                            regexp_group=1)
    assert not r[0].endswith("gz")

def test_datasets_in_custom_path(tmpdir_factory):
    """
    check that a dataset in a user-defined CTAPIPE_SVC_PATH is located
    """

    tmpdir1 = tmpdir_factory.mktemp('datasets1')
    tmpdir2 = tmpdir_factory.mktemp('datasets2')
    os.environ['CTAPIPE_SVC_PATH'] = ":".join([str(tmpdir1),str(tmpdir2)])

    # create a dummy dataset to search for:

    dataset_name = "test_dataset_1.txt"
    dataset_path = str(tmpdir1.join(dataset_name))

    with open(dataset_path, "w") as fp:
        fp.write("test test test")

    # try to find dummy dataset
    path = datasets.get_dataset(dataset_name)
    assert path == dataset_path

    with pytest.raises(FileNotFoundError):
        badpath = datasets.get_dataset("does_not_exist")


    # try using find_all_matching_datasets:

    ds = datasets.find_all_matching_datasets("test.*",
                                             searchpath=os.environ['CTAPIPE_SVC_PATH'])
    assert dataset_name in ds


from ctapipe.coordinates import *
import numpy as np
import astropy.units as u


def test_cam_to_tel():

    # Coordinates in any fram can be given as a numpy array of the xyz positions
    # e.g. in this case the position on pixels in the camera
    pix = [np.ones(1),np.zeros(1), np.zeros(1)] * u.m
    # first define the camera frame
    camera_coord = CameraFrame(pix, focal_length=15*u.m, rotation=0*u.deg)

    # then use transform to function to convert to a new system
    # making sure to give the required values for the conversion (these are not checked yet)
    telescope_coord = camera_coord.transform_to(TelescopeFrame())
    assert telescope_coord.x[0] == (1./15.)*u.rad

    # check rotation
    camera_coord = CameraFrame(pix, focal_length=15*u.m, rotation=0*u.deg)
    telescope_coord_rot = camera_coord.transform_to(TelescopeFrame())
    assert telescope_coord_rot.y[0] - (1./15.)*u.rad < 1e-6 * u.rad

    # The Transform back
    camera_coord2 = telescope_coord.transform_to(CameraFrame(focal_length=15*u.m, rotation=0*u.deg))

    # Check separation
    assert camera_coord.separation_3d(camera_coord2)[0] == 0 * u.m


def test_ground_to_tilt():

    # define ground coordinate
    grd_coord = GroundFrame(x=1*u.m, y=2*u.m, z=0*u.m)

    # Convert to tilted frame at zenith (should be the same)
    tilt_coord = grd_coord.transform_to(TiltedGroundFrame(pointing_direction=HorizonFrame(alt=90*u.deg, az=0*u.deg)))
    assert tilt_coord.separation_3d(grd_coord) == 0 * u.m

    # Check 180 degree rotation reverses y coordinate
    tilt_coord = grd_coord.transform_to(TiltedGroundFrame(pointing_direction=HorizonFrame(alt=90*u.deg, az=180*u.deg)))
    assert np.abs(tilt_coord.y + 2. * u.m) < 1e-5 * u.m

    # Check that if we look at horizon the x coordinate is 0
    tilt_coord = grd_coord.transform_to(TiltedGroundFrame(pointing_direction=HorizonFrame(alt=0*u.deg, az=0*u.deg)))
    assert np.abs(tilt_coord.x) < 1e-5 * u.m



import re
from abc import abstractmethod, ABCMeta
from collections import defaultdict
from functools import partial

import numpy as np
import tables
from astropy.time import Time
from astropy.units import Quantity

import ctapipe
from ctapipe.core import Component

__all__ = ['TableWriter',
           'TableReader',
           'HDF5TableWriter',
           'HDF5TableReader']

PYTABLES_TYPE_MAP = {
    'float': tables.Float64Col,
    'float64': tables.Float64Col,
    'float32': tables.Float32Col,
    'int': tables.IntCol,
    'int32': tables.Int32Col,
    'int64': tables.Int64Col,
    'bool': tables.BoolCol,
}


class TableWriter(Component, metaclass=ABCMeta):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._transforms = defaultdict(dict)
        self._exclusions = defaultdict(list)

    def exclude(self, table_name, pattern):
        """
        Exclude any columns matching the pattern from being written

        Parameters
        ----------
        table_name: str
            name of table on which to apply the exclusion
        pattern: str
            regular expression string to match column name
        """
        self._exclusions[table_name].append(re.compile(pattern))

    def _is_column_excluded(self, table_name, col_name):
        for pattern in self._exclusions[table_name]:
            if pattern.match(col_name):
                return True
        return False

    def add_column_transform(self, table_name, col_name, transform):
        """
        Add a transformation function for a column. This function will be
        called on the value in the container before it is written to the
        output file.

        Parameters
        ----------
        table_name: str
            identifier of table being written
        col_name: str
            name of column in the table (or item in the Container)
        transform: callable
            function that take a value and returns a new one
        """
        self._transforms[table_name][col_name] = transform
        self.log.debug("Added transform: {}/{} -> {}".format(table_name,
                                                             col_name,
                                                             transform))

    @abstractmethod
    def write(self, table_name, container):
        """
        Write the contents of the given container to a table.  The first call
        to write  will create a schema and initialize the table within the
        file. The shape of data within the container must not change between
        calls, since variable-length arrays are not supported.

        Parameters
        ----------
        table_name: str
            name of table to write to
        container: `ctapipe.core.Container`
            container to write
        """
        pass

    def _apply_col_transform(self, table_name, col_name, value):
        """
        apply value transform function if it exists for this column
        """
        if col_name in self._transforms[table_name]:
            tr = self._transforms[table_name][col_name]
            value = tr(value)
        return value


class HDF5TableWriter(TableWriter):
    """
    A very basic table writer that can take a container (or more than one)
    and write it to an HDF5 file. It does _not_ recursively write the
    container. This is intended as a building block to create a more complex
    I/O system.

    It works by creating a HDF5 Table description from the `Field`s inside a
    container, where each item becomes a column in the table. The first time
    `SimpleHDF5TableWriter.write()` is called, the container is registered
    and the table created in the output file.

    Each item in the container can also have an optional transform function
    that is called before writing to transform the value.  For example,
    unit quantities always have their units removed, or converted to a
    common unit if specified in the `Field`.

    Any metadata in the `Container` (stored in `Container.meta`) will be
    written to the table's header on the first call to write()

    Multiple tables may be written at once in a single file, as long as you
    change the table_name attribute to write() to specify which one to write
    to.

    TODO:
    - ability to write several containers to the same table (appending a
    string to each column name). Perhaps `write(name, dict(method_a=cont,
    method_b=cont2))`, where "method_a_X" would be a column name. May be
    possible to do this with some container magic, like joining two
    containers `joined_container(cont1, cont2, "A", "B")` or "cont1+cont2".
    Perhaps need to provide a better way to get container contents as a
    dictionary.

    Parameters
    ----------
    filename: str
        name of hdf5 output file
    group_name: str
        name of group into which to put all of the tables generated by this
        Writer (it will be placed under "/" in the file)

    """

    def __init__(self, filename, group_name, **kwargs):
        super().__init__()
        self._schemas = {}
        self._tables = {}
        self._h5file = tables.open_file(filename, mode="w", **kwargs)
        self._group = self._h5file.create_group("/", group_name)
        self.log.debug("h5file: {}".format(self._h5file))

    def __del__(self):
        self._h5file.close()

    def _create_hdf5_table_schema(self, table_name, container):
        """
        Creates a pytables description class for a container
        and registers it in the Writer

        Parameters
        ----------
        table_name: str
            name of table
        container: ctapipe.core.Container
            instance of an initalized container

        Returns
        -------
        dictionary of extra metadata to add to the table's header
        """

        class Schema(tables.IsDescription):
            pass

        meta = {}  # any extra meta-data generated here (like units, etc)

        # create pytables schema description for the given container
        for col_name, value in container.items():

            typename = ""
            shape = 1

            if self._is_column_excluded(table_name, col_name):
                self.log.debug("excluded column: {}/{}".format(table_name,
                                                               col_name))
                continue

            if isinstance(value, Quantity):
                req_unit = container.fields[col_name].unit
                if req_unit is not None:
                    tr = partial(tr_convert_and_strip_unit, unit=req_unit)
                    meta['{}_UNIT'.format(col_name)] = str(req_unit)
                else:
                    tr = lambda x: x.value
                    meta['{}_UNIT'.format(col_name)] = str(value.unit)

                value = tr(value)
                self.add_column_transform(table_name, col_name, tr)

            if isinstance(value, np.ndarray):
                typename = value.dtype.name
                coltype = PYTABLES_TYPE_MAP[typename]
                shape = value.shape
                Schema.columns[col_name] = coltype(shape=shape)

            if isinstance(value, Time):
                # TODO: really should use MET, but need a func for that
                Schema.columns[col_name] = tables.Float64Col()
                self.add_column_transform(table_name, col_name,
                                          tr_time_to_float)

            elif type(value).__name__ in PYTABLES_TYPE_MAP:
                typename = type(value).__name__
                coltype = PYTABLES_TYPE_MAP[typename]
                Schema.columns[col_name] = coltype()

            self.log.debug("Table {}: added col: {} type: {} shape: {}"
                           .format(table_name, col_name, typename, shape))

        self._schemas[table_name] = Schema
        meta['CTAPIPE_VERSION'] = ctapipe.__version__
        return meta

    def _setup_new_table(self, table_name, container):
        """ set up the table. This is called the first time `write()`
        is called on a new table """
        self.log.debug("Initializing table '{}'".format(table_name))
        meta = self._create_hdf5_table_schema(table_name, container)
        meta.update(container.meta)  # copy metadata from container

        table = self._h5file.create_table(where=self._group,
                                          name=table_name,
                                          title="storage of {}".format(
                                              container.__class__.__name__),
                                          description=self._schemas[table_name])
        for key, val in meta.items():
            table.attrs[key] = val

        self._tables[table_name] = table

    def _append_row(self, table_name, container):
        """
        append a row to an already initialized table. This is called
        automatically by `write()`
        """
        table = self._tables[table_name]
        row = table.row

        for colname in table.colnames:
            value = self._apply_col_transform(table_name, colname,
                                              container[colname])

            row[colname] = value

        row.append()

    def write(self, table_name, container):
        """
        Write the contents of the given container to a table.  The first call
        to write  will create a schema and initialize the table within the
        file. The shape of data within the container must not change between
        calls, since variable-length arrays are not supported.

        Parameters
        ----------
        table_name: str
            name of table to write to
        container: `ctapipe.core.Container`
            container to write
        """

        if table_name not in self._schemas:
            self._setup_new_table(table_name, container)

        self._append_row(table_name, container)


class TableReader(Component, metaclass=ABCMeta):
    """
    Base class for row-wise table readers. Generally methods that read a
    full table at once are preferred to this method, since they are faster,
    but this can be used to re-play a table row by row into a
    `ctapipe.core.Container` class (the opposite of TableWriter)
    """

    def __init__(self):
        super().__init__()
        self._cols_to_read = defaultdict(list)
        self._transforms = defaultdict(dict)

    def add_column_transform(self, table_name, col_name, transform):
        """
        Add a transformation function for a column. This function will be
        called on the value in the container before it is written to the
        output file.

        Parameters
        ----------
        table_name: str
            identifier of table being written
        col_name: str
            name of column in the table (or item in the Container)
        transform: callable
            function that take a value and returns a new one
        """
        self._transforms[table_name][col_name] = transform
        self.log.debug("Added transform: {}/{} -> {}".format(table_name,
                                                             col_name,
                                                             transform))

    def _apply_col_transform(self, table_name, col_name, value):
        """
        apply value transform function if it exists for this column
        """
        if col_name in self._transforms[table_name]:
            tr = self._transforms[table_name][col_name]
            value = tr(value)
        return value

    @abstractmethod
    def read(self, table_name, container):
        """
        Returns a generator that reads the next row from the table into the
        given container.  The generator returns the same container. Note that
        no containers are copied, the data are overwritten inside.
        """
        pass


class HDF5TableReader(TableReader):
    """
    Reader that reads a single row of an HDF5 table at once into a Container.
    Simply construct a `HDF5TableReader` with an input HDF5 file,
    and call the `read(path, container)` method to get a generator that fills
    the given container with a new row of the table on each access.

    Columns in the table are automatically mapped to container fields by
    name, and if a field is missing in either, it is skipped during read,
    but a warning is emitted.

    Columns that were written by SimpleHDF5TableWriter and which had unit
    transforms applied, will have the units re-applied when reading (the
    unit used is stored in the header attributes).

    Note that this is only useful if you want to read all information *one
    event at a time* into a container, which is not very I/O efficient. For
    some other use cases, it may be much more efficient to access the
    table data directly, for example to read an entire column or table at
    once (which means not using the Container data structure).

    Todo:
    - add ability to synchronize reading of multiple tables on a key

    - add ability (also with TableWriter) to read a row into n containers at
        once, assuming no naming conflicts (so we can add e.g. event_id)

    """

    def __init__(self, filename):
        """
        Parameters
        ----------
        filename: str
            name of hdf5 file
        group_name: str
            HDF5 path to group where tables are  to be found
        """
        super().__init__()
        self._tables = {}
        self._h5file = tables.open_file(filename)
        pass

    def _setup_table(self, table_name, container):
        tab = self._h5file.get_node(table_name)
        self._tables[table_name] = tab
        self._map_table_to_container(table_name, container)
        self._map_transforms_from_table_header(table_name)
        return tab

    def _map_transforms_from_table_header(self, table_name):
        """
        create any transforms needed to "undo" ones in the writer
        """
        tab = self._tables[table_name]
        for attr in tab.attrs._f_list():
            if attr.endswith("_UNIT"):
                colname = attr[:-5]
                tr = partial(tr_add_unit, unitname=tab.attrs[attr])
                self.add_column_transform(table_name, colname, tr)

    def _map_table_to_container(self, table_name, container):
        """ identifies which columns in the table to read into the container,
        by comparing their names."""
        tab = self._tables[table_name]
        for colname in tab.colnames:
            if colname in container.fields:
                self._cols_to_read[table_name].append(colname)
            else:
                self.log.warn("Table '{}' has column '{}' that is not in "
                              "container {}. It will be skipped"
                              .format(table_name, colname,
                                      container.__class__.__name__))

        # also check that the container doesn't have fields that are not
        # in the table:
        for colname in container.fields:
            if colname not in self._cols_to_read[table_name]:
                self.log.warn("Table '{}' is missing column '{}' that is "
                              "in container {}. It will be skipped"
                              .format(table_name, colname,
                                      container.__class__.__name__))

        # copy all user-defined attributes back to Container.mets
        for key in tab.attrs._f_list():
            container.meta[key] = tab.attrs[key]

    def read(self, table_name, container):
        """
        Returns a generator that reads the next row from the table into the
        given container.  The generator returns the same container. Note that
        no containers are copied, the data are overwritten inside.
        """
        if table_name not in self._tables:
            tab = self._setup_table(table_name, container)
        else:
            tab = self._tables[table_name]


        row_count = 0

        while 1:

            try:
                row = tab[row_count]
            except IndexError:
                return # stop generator when done

            for colname in self._cols_to_read[table_name]:
                container[colname] = self._apply_col_transform(table_name,
                                                               colname,
                                                               row[colname])

            yield container
            row_count += 1






def tr_convert_and_strip_unit(quantity, unit):
    return quantity.to(unit).value


def tr_list_to_mask(thelist, length):
    """ transform list to a fixed-length mask"""
    arr = np.zeros(shape=length, dtype=np.bool)
    arr[thelist] = True
    return arr


def tr_time_to_float(thetime):
    return thetime.mjd


def tr_add_unit(value, unitname):
    return Quantity(value, unitname)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Components to read ZFITS data.

This requires the protozfitsreader python library to be installed
"""
import logging

import numpy as np
import numpy.ma as ma

from .containers import R0Container, R0CameraContainer
from ctapipe.core import Container

from astropy import units as u
from astropy.coordinates import Angle
from astropy.time import Time

logger = logging.getLogger(__name__)

try:
    import protozfitsreader
except ImportError as err:
    logger.fatal("the `protozfitsreader` python module is required to access MC data: {}"
                 .format(err))
    raise err

__all__ = [
    'zfits_event_source',
]

def zfits_event_source(url, max_events=None, allowed_tels=None):
    """A generator that streams data from an EventIO/HESSIO MC data file
    (e.g. a standard CTA data file.)

    Parameters
    ----------
    url : str
        path to file to open
    max_events : int, optional
        maximum number of events to read
    allowed_tels : list[int]
        select only a subset of telescope, if None, all are read. This can
        be used for example emulate the final CTA data format, where there
        would be 1 telescope per file (whereas in current monte-carlo,
        they are all interleaved into one file)

    """

    try:
        zfits = protozfitsreader.ZFile(url)
    except:
        raise RuntimeError("zfits_event_source failed to open '{}'"
                           .format(url))


    counter=0
    eventstream = zfits.move_to_next_event()

    container = Container("zfits_container")
    container.meta.add_item('zfits__input', url)
    container.meta.add_item('zfits__max_events', events)
    container.meta.add_item('pixel_pos', dict())
    container.add_item("r0", R0Container())
    container.add_item("trig", CentralTriggerData())
    container.add_item("count")

    for run_id, event_id in eventstream:
        container.dl0.run_id    = run_id
        container.dl0.event_id  = event_id
        #container.r0.event_num = zfits.get_event_number()

        # We assume we are in the single-telescope case always:
        container.dl0.tels_with_data     = [zfits.get_telescope_id(), ]
        container.trig.tels_with_trigger = container.dl0.tels_with_data

        time_s, time_ns = zfits.get_central_event_gps_time()
        container.trig.gps_time = Time(time_s * u.s, time_ns * u.ns,
                                       format='gps', scale='utc')

        container.count = counter
        t = np.arange(n_samples)
        
        container.dl0.tel = dict()  # clear the previous telescopes

        # Depecrated loop, we keep it for clarity (similar structure than hessio and toymodel modules)
        for tel_id in container.dl0.tels_with_data:
            # fill pixel position dictionary, if not already done:
            #TODO: tel_id here is a dummy parameter, we are dealing with single-telescope data!. TBR.
            if tel_id not in container.inst.pixel_pos:
                container.inst.pixel_pos[tel_id] = \
                    zfits.get_pixel_position(tel_id) * u.m

            nchans = zfits.get_num_channels(tel_id)
            container.inst.num_channels[tel_id] = nchans
            
            for chan in range(nchans): 
                samples = zfits.get_adc_sample(channel=chan, telescope_id=tel_id)
                integrated = zfits.get_adc_sum(channel=chan, telescope_id=tel_id)
                
                container.dl0.tel[tel_id].pixel_samples[chan] = samples.keys()

                mask = np.zeros(zfits.get_number_of_pixels(),dtype=bool)
                mask[np.array(samples.keys())]=True
                container.dl0.tel[tel_id].adc_samples[chan] = \
                    ma.array(np.array(samples.values()),mask=mask)
                
                mask = np.zeros(zfits.get_number_of_pixels(),dtype=bool)
                mask[np.array(integrated.keys())]=True
                container.dl0.tel[tel_id].adc_integrated[chan] = \
                    ma.array(np.array(integrated.values()),mask=mask)

        yield container
        counter +=1

        if max_events is not None and counter > max_events:
            return




"""
This module defines any reference systems which may be needed in addition
"""

from astropy.coordinates import BaseRepresentation, CartesianRepresentation
import astropy.units as u
from collections import OrderedDict
from astropy.utils.compat.numpy import broadcast_arrays


class PlanarRepresentation(BaseRepresentation):
    """
    Representation of a point in a 2D plane. This is essentially a copy of the Cartesian representation used
    in astropy.

    Parameters
    ----------

    x, y : `~astropy.units.Quantity`
        The x and y coordinates of the point(s). If ``x`` and ``y``have
        different shapes, they should be broadcastable.

    copy : bool, optional
        If True arrays will be copied rather than referenced.

    """
    attr_classes = OrderedDict([('x', u.Quantity),
                                ('y', u.Quantity)])

    def __init__(self, x, y, copy=True, **kwargs):


        if x is None or y is None:
            raise ValueError("x and y are required to instantiate CartesianRepresentation")

        if not isinstance(x, self.attr_classes['x']):
            raise TypeError('x should be a {0}'.format(self.attr_classes['x'].__name__))

        if not isinstance(y, self.attr_classes['y']):
            raise TypeError('y should be a {0}'.format(self.attr_classes['y'].__name__))

        x = self.attr_classes['x'](x, copy=copy)
        y = self.attr_classes['y'](y, copy=copy)

        if not (x.unit.physical_type == y.unit.physical_type):
            raise u.UnitsError("x and y should have matching physical types")

        try:
            x, y = broadcast_arrays(x, y, subok=True)
        except ValueError:
            raise ValueError("Input parameters x and y cannot be broadcast")

        self._x = x
        self._y = y
        self._differentials = {}


    @property
    def x(self):
        """
        The x component of the point(s).
        """
        return self._x

    @property
    def y(self):
        """
        The y component of the point(s).
        """
        return self._y

    @property
    def xy(self):
        return u.Quantity((self._x, self._y))

    @classmethod
    def from_cartesian(self, cartesian):

        return PlanarRepresentation(x=cartesian.x, y=cartesian.y)

    def to_cartesian(self):

        return CartesianRepresentation(x=self._x, y=self._y, z=0*self._x.unit)

    @property
    def components(self):
        return 'x', 'y'



""" Class to handle configuration for algorithms """

from traitlets.config import Configurable
from abc import ABCMeta
from logging import getLogger


class AbstractConfigurableMeta(type(Configurable), ABCMeta):
    '''
    Metaclass to be able to make Component abstract
    see: http://stackoverflow.com/a/7314847/3838691
    '''
    pass


class Component(Configurable, metaclass=AbstractConfigurableMeta):
    """Base class of all Components (sometimes called
    workers, makers, etc).  Components are are classes that do some sort
    of processing and contain user-configurable parameters, which are
    implemented using `traitlets`.

    `traitlets` can validate values and provide defaults and
    descriptions. These will be automatically translated into
    configuration parameters (command-line, config file, etc). Note
    that any parameter that should be externally configurable must
    have its `config` attribute set to `True`, e.g. defined like
    `myparam = Integer(0, help='the parameter').tag(config=True)`.

    All components also contain a `logger` instance in their `log`
    attribute, that you must use to output info, debugging data,
    warnings, etc (do not use `print()` statements, instead use
    `self.log.info()`, `self.log.warn()`, `self.log.debug()`, etc).

    Components are generally used within `ctapipe.core.Tool`
    subclasses, which provide configuration handling and command-line
    tool generation.


    For example:

    .. code:: python

        from ctapipe.core import Component
        from traitlets import (Integer, Float)

        class MyComponent(Component):
            \"\"\" Does something \"\"\"
            some_option = Integer(default_value=6,
                                  help='a value to set').tag(config=True)


        comp = MyComponent(None)
        comp.some_option = 6      # ok
        comp.some_option = 'test' # will fail validation
    """

    def __init__(self, parent=None, **kwargs):
        """
        Parameters
        ----------
        parent: Tool or Component
            Tool or component that is the Parent of this one
        kwargs: type
            other paremeters

        """

        super().__init__(parent=parent, **kwargs)

        # set up logging
        if self.parent:
            self.log = self.parent.log.getChild(self.__class__.__name__)
        else:
            self.log = getLogger(self.__class__.__name__)



from tempfile import TemporaryDirectory

import numpy as np
from numpy.testing import assert_allclose

from astropy import units as u

from ctapipe.reco.energy_regressor import *


def test_prepare_model():
    cam_id_list = ["FlashCam", "ASTRICam"]
    feature_list = {"FlashCam": [[1, 10], [2, 20], [3, 30], [0.9, 9],
                                 ],
                    "ASTRICam": [[10, 1], [20, 2], [30, 3], [9, 0.9],
                                 ]}
    target_list = {"FlashCam": np.array([1, 2, 3, 0.9])*u.TeV,
                   "ASTRICam": np.array([1, 2, 3, 0.9])*u.TeV}

    reg = EnergyRegressor(cam_id_list=cam_id_list)
    reg.fit(feature_list, target_list)
    return reg, cam_id_list


def test_fit_save_load():
    reg, cam_id_list = test_prepare_model()
    with TemporaryDirectory() as d:
        temp_path = "/".join([d, "reg_{cam_id}.pkl"])
        reg.save(temp_path)
        reg = EnergyRegressor.load(temp_path, cam_id_list)
        return reg, cam_id_list


def test_predict_by_event():
    np.random.seed(3)

    reg, cam_id_list = test_fit_save_load()
    prediction = reg.predict_by_event([{"ASTRICam": [[10, 1]]},
                                       {"ASTRICam": [[20, 2]]},
                                       {"ASTRICam": [[30, 3]]}])
    assert_allclose(prediction["mean"].value, [1, 2, 3], rtol=0.2)

    prediction = reg.predict_by_event([{"FlashCam": [[1, 10]]},
                                       {"FlashCam": [[2, 20]]},
                                       {"FlashCam": [[3, 30]]}])
    assert_allclose(prediction["mean"].value, [1, 2, 3], rtol=0.2)



from ctapipe.utils.CutFlow import CutFlow, \
                                  UndefinedCutException, \
                                  PureCountingCutException
import numpy as np
from pytest import raises


def test_CutFlow():
    flow = CutFlow("TestFlow")
    # set_cut and add_cut a aliases
    flow.set_cut("smaller5", lambda x: x < 5)
    flow.add_cut("smaller3", lambda x: x < 3)

    for i in range(2, 6):
        flow.count("noCuts")
        # .keep counts if the function returns True,
        # i.e. when we "keep" the event
        if flow.keep("smaller5", i):
            # .cut counts if the function returns False,
            # i.e. when we do NOT "cut" the event
            if flow.cut("smaller3", i):
                pass
            else:
                # do something else that could fail or be rejected
                try:
                    assert i == 3
                    flow.count("something")
                except:
                    pass


    t = flow(sort_column=1)
    assert np.all(t["selected Events"] == [4, 3, 2, 1])

    with raises(UndefinedCutException):
        flow.cut("undefined", 5)

    with raises(PureCountingCutException):
        flow.cut("noCuts")


if __name__ == "__main__":
    test_CutFlow()





import numpy as np
from numpy.testing import assert_almost_equal

from ctapipe.image.charge_extractors import FullIntegrator, \
    SimpleIntegrator, GlobalPeakIntegrator, LocalPeakIntegrator, \
    NeighbourPeakIntegrator, ChargeExtractorFactory, AverageWfPeakIntegrator
from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset


def get_test_event():
    filename = get_dataset('gamma_test.simtel.gz')
    source = hessio_event_source(filename, requested_event=409,
                                 use_event_id=True)
    event = next(source)
    return event


def test_full_integration():
    telid = 11
    event = get_test_event()
    data = event.r0.tel[telid].adc_samples
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped/nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = FullIntegrator(None, None)
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 149, 0)
    assert_almost_equal(integration[1][0], 149, 0)
    assert peakpos[0][0] == 0
    assert peakpos[1][0] == 0


def test_simple_integration():
    telid = 11
    event = get_test_event()
    data = event.r0.tel[telid].adc_samples
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped/nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = SimpleIntegrator(None, None)
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 74, 0)
    assert_almost_equal(integration[1][0], 74, 0)
    assert peakpos[0][0] == 0
    assert peakpos[1][0] == 0


def test_global_peak_integration():
    telid = 11
    event = get_test_event()
    data = event.r0.tel[telid].adc_samples
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped/nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = GlobalPeakIntegrator(None, None)
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 58, 0)
    assert_almost_equal(integration[1][0], 58, 0)
    assert peakpos[0][0] == 14
    assert peakpos[1][0] == 14


def test_local_peak_integration():
    telid = 11
    event = get_test_event()
    data = event.r0.tel[telid].adc_samples
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped/nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = LocalPeakIntegrator(None, None)
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 76, 0)
    assert_almost_equal(integration[1][0], 76, 0)
    assert peakpos[0][0] == 13
    assert peakpos[1][0] == 13


def test_nb_peak_integration():
    telid = 11
    event = get_test_event()
    data = event.r0.tel[telid].adc_samples
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped/nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality
    geom = event.inst.subarray.tel[telid].camera
    nei = geom.neighbors

    integrator = NeighbourPeakIntegrator(None, None)
    integrator.neighbours = nei
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], -64, 0)
    assert_almost_equal(integration[1][0], -64, 0)
    assert peakpos[0][0] == 20
    assert peakpos[1][0] == 20


def test_averagewf_peak_integration():
    telid = 11
    event = get_test_event()
    data = event.r0.tel[telid].adc_samples
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped/nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = AverageWfPeakIntegrator(None, None)
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 73, 0)
    assert_almost_equal(integration[1][0], 73, 0)
    assert peakpos[0][0] == 10
    assert peakpos[1][0] == 10


def test_charge_extractor_factory():
    extractor_f = ChargeExtractorFactory(None, None)
    extractor_f.extractor = 'LocalPeakIntegrator'
    extractor_c = extractor_f.get_class()
    extractor = extractor_c(None, None)

    telid = 11
    event = get_test_event()
    data = event.r0.tel[telid].adc_samples
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped/nsamples)

    integration, peakpos, window = extractor.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 76, 0)



"""
Class for calculation of likelihood of a pixel expectation, given the pixel amplitude,
the level of noise in the pixel and the photoelectron resolution. This calculation is
taken from:
de Naurois & Rolland, Astroparticle Physics, Volume 32, Issue 5, p. 231-252 (2009)
https://arxiv.org/abs/0907.2610

The likelihood is essentially a poissonian convolved with a gaussian, at low signal
a full possonian approach must be adopted, which requires the sum of contibutions
over a number of potential contributing photoelectrons (which is slow and can fail
at high signals due to the factorial which mst be calculated). At high signal this
simplifies to a gaussian approximation.

The full and gaussian approximations are implemented, in addition to a general purpose
implementation, which tries to intellegently switch between the two. Speed tests are below:

poisson_likelihood_gaussian(image, prediction, spe, ped)
29.8 µs per loop

poisson_likelihood_full(image, prediction, spe, ped)
93.4 µs per loop

poisson_likelihood(image, prediction, spe, ped)
59.9 µs per loop

TODO:
=====
- Need to implement more tests, particularly checking for error states
- Additional terms may be useful to add to the likelihood
"""

import math

import numpy as np
from scipy.integrate import quad
from scipy.misc import factorial

__all__ = [
    'poisson_likelihood_gaussian',
    'poisson_likelihood_full',
    'poisson_likelihood',
    'mean_poisson_likelihood_gaussian',
    'mean_poisson_likelihood_full',
    'PixelLikelihoodError',
    'chi_squared'
]


class PixelLikelihoodError(RuntimeError):
    pass


def poisson_likelihood_gaussian(image, prediction, spe_width, ped):
    """
    Calculate likelihood of prediction given the measured signal, gaussian approx from
    de Naurois et al 2009

    Parameters
    ----------
    image: ndarray
        Pixel amplitudes from image
    prediction: ndarray
        Predicted pixel amplitudes from model
    spe_width: ndarray
        width of single p.e. distributio
    ped: ndarray
        width of pedestal

    Returns
    -------
    ndarray: likelihood for each pixel
    """
    image = np.asarray(image)
    prediction = np.asarray(prediction)
    spe_width = np.asarray(spe_width)
    ped = np.asarray(ped)

    sq = 1. / np.sqrt(2 * math.pi * (np.power(ped, 2)
                                     + prediction * (1 + np.power(spe_width, 2))))

    diff = np.power(image - prediction, 2.)
    denom = 2 * (np.power(ped, 2) + prediction * (1 + np.power(spe_width, 2)))
    expo = np.asarray(np.exp(-1 * diff / denom))

    # If we are outside of the range of datatype, fix to lower bound
    min_prob = np.finfo(expo.dtype).tiny
    expo[expo < min_prob] = min_prob

    return -2 * np.log(sq * expo)


def poisson_likelihood_full(
        image, prediction, spe_width, ped, width_fac=3, dtype=np.float32):
    """
    Calculate likelihood of prediction given the measured signal, full numerical integration from
    de Naurois et al 2009. The width factor included here defines the range over which photo electron contributions
    are summed, and is defined as a multiple of the expected resolution of the highest amplitude pixel. For
    most applications the defult of 3 is sufficient.

    Parameters
    ----------
    image: ndarray
        Pixel amplitudes from image
    prediction: ndarray
        Predicted pixel amplitudes from model
    spe_width: ndarray
        width of single p.e. distribution
    ped: ndarray
        width of pedestal
    width_fac: float
        Factor to determine range of summation on integral
    dtype: datatype
        Data type of output array

    Returns
    -------
    ndarray: likelihood for each pixel
    """

    image = np.asarray(image, dtype=dtype)
    prediction = np.asarray(prediction, dtype=dtype)
    spe_width = np.asarray(spe_width, dtype=dtype)
    ped = np.asarray(ped, dtype=dtype)

    if image.shape != prediction.shape:
        raise PixelLikelihoodError("Image and prediction arrays have different dimensions",
                                   "Image shape: ", image.shape[0], "Prediction shape: ", prediction.shape[0])
    max_val = np.max(image)
    width = ped * ped + max_val * spe_width * spe_width
    width = np.sqrt(np.abs(width))  # take abs of width if negative

    max_sum = max_val + width_fac * width
    if max_sum < 5:
        max_sum = 5

    pe_summed = np.arange(max_sum)  # Need to decide how range is determined
    pe_factorial = factorial(pe_summed)

    first_term = np.power(prediction, pe_summed[
                          :, np.newaxis]) * np.exp(-1 * prediction)
    first_term /= pe_factorial[:, np.newaxis] * \
        np.sqrt(math.pi * 2 * (ped * ped +
                               pe_summed[:, np.newaxis] * spe_width * spe_width))

    # Throw error if we get NaN in likelihood
    if np.any(np.isnan(first_term)):
        raise PixelLikelihoodError("Likelihood returning NaN, likely due to extremely high signal"
                                   " deviation. Switch to poisson_likelihood_safe implementation or"
                                   " increase floating point precision e.g. dtype=float64")

    # Should not have any porblems here with NaN that have not bee seens
    second_term = (image - pe_summed[:, np.newaxis]) * \
        (image - pe_summed[:, np.newaxis])
    second_term_denom = 2 * (ped * ped + spe_width *
                             spe_width * pe_summed[:, np.newaxis])

    second_term = second_term / second_term_denom
    second_term = np.exp(-1 * second_term)

    # If we are outside of the range of datatype, fix to lower bound
    min_prob = np.finfo(second_term.dtype).tiny
    second_term[second_term < min_prob] = min_prob

    like = first_term * second_term

    return -2 * np.log(np.sum(like, axis=0))


def poisson_likelihood(image, prediction, spe_width, ped,
                       pedestal_safety=1.5, width_fac=3, dtype=np.float32):
    """
    Safe implementation of the poissonian likelihood implementation , adaptively switches between the full
    solution and the gaussian approx depending on the signal. Pedestal safety parameter determines cross
    over point between the two solutions, based on the expected p.e. resolution of the image pixels. Therefore
    the cross over point will change dependent on the single p.e. resolution and pedestal levels.

    Parameters
    ----------
    image: ndarray
        Pixel amplitudes from image
    prediction: ndarray
        Predicted pixel amplitudes from model
    spe_width: ndarray
        width of single p.e. distribution
    ped: ndarray
        width of pedestal
    pedestal_safety: float
        Decision point to choose between poissonian likelihood and gaussian approximation (p.e. resolution)
    width_fac: float
        Factor to determine range of summation on integral
    dtype: datatype
        Data type of output array

    Returns
    -------
    ndarray: pixel likelihoods
    """
    # Convert everything to arrays to begin
    image = np.asarray(image, dtype=dtype)
    prediction = np.asarray(prediction, dtype=dtype)
    spe_width = np.asarray(spe_width, dtype=dtype)
    ped = np.asarray(ped, dtype=dtype)

    # Calculate photoelectron resolution

    width = ped * ped + image * spe_width * spe_width
    width=np.asarray(width)
    width[width < 0] = 0  # Set width to 0 for negative pixel amplitudes
    width = np.sqrt(width)

    like = np.zeros(image.shape)
    # If larger than safety value use gaussian approx
    poisson_pix = width <= pedestal_safety
    gaus_pix = width > pedestal_safety

    if np.any(poisson_pix):
        like[poisson_pix] = poisson_likelihood_full(image[poisson_pix], prediction[poisson_pix],
                                                    spe_width, ped, width_fac, dtype)
    if np.any(gaus_pix):
        like[gaus_pix] = poisson_likelihood_gaussian(image[gaus_pix], prediction[gaus_pix],
                                                     spe_width, ped)

    return like


def mean_poisson_likelihood_gaussian(prediction, spe_width, ped):
    """
    Calculation of the mean  likelihood for a give expectation value of pixel intensity in the gaussian approximation.
    This is useful in the calculation of the goodness of fit.

    Parameters
    ----------
    prediction: ndarray
        Predicted pixel amplitudes from model
    spe_width: ndarray
        width of single p.e. distribution
    ped: ndarray
        width of pedestal

    Returns
    -------
    ndarray: mean likelihood for give pixel expectation
    """
    prediction = np.asarray(prediction)
    spe_width = np.asarray(spe_width)
    ped = np.asarray(ped)

    mean_like = 1 + np.log(2 * math.pi)
    mean_like += np.log(ped * ped + prediction * (1 + spe_width * spe_width))

    return mean_like


def _integral_poisson_likelihood_full(s, prediction, spe_width, ped):
    """
    Wrapper function around likelihood calculation, used in numerical integration.
    """
    like = poisson_likelihood(s, prediction, spe_width, ped)
    return like * np.exp(-0.5 * like)


def mean_poisson_likelihood_full(prediction, spe_width, ped):
    """
    Calculation of the mean  likelihood for a give expectation value of pixel intensity using the full
    numerical integration. This is useful in the calculation of the goodness of fit. This numerical integration
    is very slow and really doesn't make a large difference in the goodness of fit in most cases.

    Parameters
    ----------
    prediction: ndarray
        Predicted pixel amplitudes from model
    spe_width: ndarray
        width of single p.e. distribution
    ped: ndarray
        width of pedestal

    Returns
    -------
    ndarray: mean likelihood for give pixel expectation
    """
    prediction = np.asarray(prediction)
    spe_width = np.asarray(spe_width)

    if len(spe_width.shape) == 0:
        spe_width = np.ones(prediction.shape) * spe_width
    ped = np.asarray(ped)
    if len(ped.shape) == 0:
        ped = np.ones(prediction.shape) * ped
    mean_like = np.zeros(prediction.shape)
    width = ped * ped + prediction * spe_width * spe_width
    width = np.sqrt(width)

    for p in range(len(prediction)):
        int_range = (prediction[p] - 10 * width[p],
                     prediction[p] + 10 * width[p])
        mean_like[p] = quad(_integral_poisson_likelihood_full, int_range[0], int_range[1],
                            args=(prediction[p], spe_width[p], ped[p]), epsrel=0.05)[0]
    return mean_like


def chi_squared(image, prediction, ped, error_factor=2.9):
    """
    Simple chi-squared statistic from Le Bohec et al 2008

    Parameters
    ----------
    image: ndarray
        Pixel amplitudes from image
    prediction: ndarray
        Predicted pixel amplitudes from model
    ped: ndarray
        width of pedestal
    error_factor: float
        ad hoc error factor

    Returns
    -------
    ndarray: likelihood for each pixel
    """

    image = np.asarray(image)
    prediction = np.asarray(prediction)
    ped = np.asarray(ped)

    if image.shape is not prediction.shape:
        PixelLikelihoodError("Image and prediction arrays have different dimensions",
                             "Image shape: ", image.shape, "Prediction shape: ", prediction.shape)

    chi_square = (image - prediction) * (image - prediction)
    chi_square /= ped + 0.5 * (image - prediction)
    chi_square *= 1. / error_factor

    return chi_square



import numpy as np
from ctapipe.image import poisson_likelihood_full, poisson_likelihood_gaussian

def test_full_likelihood():
    """
    Simple test of likelihood, test against known values for high and low 
    signal cases. Check that full calculation and the gaussian approx become 
    equal at high signal.
    """
    spe = 0.5 # Single photo-electron width
    pedestal = 1 # width of the pedestal distribution

    image_small = [0,1,2]
    expectation_small = [1,1,1]

    full_like_small = poisson_likelihood_full(image_small, expectation_small,
                                              spe, pedestal)
    exp_diff = full_like_small - np.asarray([2.75630505, 2.62168656, 3.39248449])
    exp_diff = np.sum(np.abs(exp_diff))
    # Check against known values
    print(exp_diff)
    assert exp_diff/np.sum(full_like_small) < 1e-4

    image_large = [40,50,60]
    expectation_large = [50,50,50]
    full_like_large = poisson_likelihood_full(image_large, expectation_large,
                                              spe, pedestal)
    # Check against known values
    exp_diff = full_like_large - np.asarray([7.45489137,5.99305388,7.66226007])
    exp_diff = np.sum(np.abs(exp_diff))

    assert exp_diff/np.sum(full_like_large) < 1e-4

    gaus_like_large = poisson_likelihood_gaussian(image_large,
                                                  expectation_large,
                                                  spe,
                                                  pedestal)

    # Check thats in large signal case the full expectation is equal to the
    # gaussian approximation (to 5%)
    assert np.all(np.abs((full_like_large-gaus_like_large)/full_like_large)
                  < 0.05)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Array layout utilities
"""
from glob import glob
from astropy.table import Table
from ctapipe.utils import get_dataset

_telclass_map = {0: 'SST', 1: 'MST', 2: 'LST'}


def get_array_layout(instrument_name):
    """
    Returns the array layout for the given instrument as an
    `astropy.table.Table` object. 
    """
    name = instrument_name.lower()
    try:
        layoutfile = get_dataset('{}_arraylayout.fits'.format(name))
    except KeyError:
        layoutfile = get_dataset('{}_arraylayout.fits.gz'.format(name))
    return load_array_layout_from_file(layoutfile)


def load_array_layout_from_file(array_layout_filename):
    """
    Read an array layout from a FITS file with a ``TELARRAY`` extension
    """
    return Table.read(array_layout_filename, hdu='TELARRAY')


def tel_class_name(tel_class):
    """Convert telescope class number to a string.
    """
    return _telclass_map[int(tel_class)]



from graphviz import Digraph
from PyQt4.QtGui import QWidget
from PyQt4.QtGui import QPainter
from PyQt4.QtSvg import QSvgRenderer
from ctapipe.flow.stager_rep import StagerRep

class GraphWidget(QWidget):

    """
    class that displays pipeline workload.
    It receives pipeline information thanks to pipechange method
    """
    def __init__(self, statusBar):
        super(GraphWidget, self).__init__()
        self.initUI()
        # self.steps contains all pipeline steps (Producer, Stagers, consumer)
        self.steps = list()
        self.statusBar = statusBar
        self.last_diagram_bytes = None

    def initUI(self):
        self.show()

    def paintEvent(self, e):
        """This event handler is reimplemented in this subclass to receive
        paint events passed in event. A paint event is a request to repaint all
        or part of a widget. It can happen for one of the following reasons:
        repaint() or update() was invoked,the widget was obscured and has
        now been uncovered, or many other reasons. """
        qp = QPainter()
        qp.begin(self)
        # if no steos have been receive draw last_diagram_bytes
        if not self.steps and self.last_diagram_bytes:
            diagram_bytes = self.last_diagram_bytes
        else:
            diagram = self.build_graph()
            diagram_bytes = diagram.pipe('svg')
            self.last_diagram_bytes = diagram_bytes
        svg = QSvgRenderer(diagram_bytes)
        svg.render(qp)
        qp.end()

    def pipechange(self, steps):
        """
        Called by GuiConnexion instance when it receives zmq message from Flow.
        Update pipeline state (self.steps) and force to update drawing

        Parameters
        ----------
        steps: list of (StagerRep)
        """
        if steps:
            self.steps = steps
            self.update()

    def reset(self):
        """Clear self.steps and self.last_diagram_bytes
        """
        self.steps = list()
        self.last_diagram_bytes = None

    def build_graph(self):
        """
        Returns
        -------
        graphiz.Digraph
            It contains nodes and links corresponding to self.steps
        """
        g = Digraph('test', format='svg',graph_attr={'bgcolor':'lightgrey'})
        #Create nodes
        for step in  self.steps:
            str_shape = 'octagon'
            if step.type == StagerRep.CONSUMER:
                str_shape = 'doubleoctagon'
            if step.type == StagerRep.PRODUCER:
                str_shape = 'Mdiamond'
            name = step.name.split('$$process')[0]
            name = self.format_name(step.name.split('$$process')[0])
            if step.running > 0:
                g.node(name,color='lightblue', style='filled',shape=str_shape,area='0.5')
            else:
                g.node(name,shape=str_shape,color='blue',area='0.5')
        #Create edges
        for step in self.steps:
            step_name = self.format_name(step.name.split('$$process')[0])
            for next_step_name in step.next_steps:
                next_step = self.get_step_by_name(next_step_name.split('$$process')[0])
                if next_step:
                    next_step_name_formated = self.format_name(next_step.name.split('$$process')[0])
                    g.edge(step_name, next_step_name_formated)
                    g.edge_attr.update(arrowhead='empty', arrowsize='1',color='purple')
        return g

    def format_name(self,name, max_car=15):
        """
        trim name if its length is more than max_car and add 3 points

        Returns
        -------
        Name triped or not
        """
        if len(name) > max_car:
            name = name[0:max_car] + '...'
        return name

    def get_step_by_name(self, name):
        ''' Find a PipeStep in self.producer_step or  self.stager_steps or
        self.consumer_step

        Parameters
        ----------
        name : str Step name

        Returns
        -------
        PipeStep if found, otherwise None
        '''
        for step in self.steps:
            if step.name.split('$$process')[0] == name:
                return step
        return None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .fitshistogram import *
from .json2fits import *
from .dynamic_class import *
from .table_interpolator import TableInterpolator
from .datasets import *



"""
Algorithms for the data volume reduction.
"""

from abc import abstractmethod
from traitlets import CaselessStrEnum
from ctapipe.core import Component, Factory

__all__ = ['DataVolumeReductor', 'DataVolumeReductorFactory']


class DataVolumeReductor(Component):
    """
    Base component for data volume reductors.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    kwargs

    Attributes
    ----------
    extracted_samples : ndarray
        Numpy array containing True where the samples lay within the
        integration window, and False where the samples lay outside. Has
        shape of (n_chan, n_pix, n_samples).
    peakpos : ndarray
        Numpy array of the peak position for each pixel. Has shape of
        (n_chan, n_pix).
    neighbours : list
        List of neighbours for each pixel. Changes per telescope.

    """
    name = 'DataVolumeReductor'

    def __init__(self, config, tool, **kwargs):
        super().__init__(config=config, parent=tool, **kwargs)

        self._nchan = None
        self._npix = None
        self._nsamples = None

        self.extracted_samples = None
        self.peakpos = None
        self.neighbours = None

    @staticmethod
    def requires_neighbours():
        """
        Method used for callers of the ChargeExtractor to know if the extractor
        requires knowledge of the pixel neighbours

        Returns
        -------
        bool
        """
        return False

    def check_neighbour_set(self):
        """
        Check if the pixel neighbours has been set for the extractor

        Raises
        -------
        ValueError
            If neighbours has not been set
        """
        if self.requires_neighbours():
            if self.neighbours is None:
                self.log.exception("neighbours attribute must be set")
                raise ValueError()

    @abstractmethod
    def reduce_waveforms(self, waveforms):
        """
        Call the relevant functions to reduce the waveforms using a
        particular reductor.

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).

        Returns
        -------
        reduced_waveforms : ndarray
            Reduced waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).
        """


class DataVolumeReductorFactory(Factory):
    """
    Factory class for creating a DataVolumeReductor
    """
    name = "DataVolumeReductorFactory"
    description = "Obtain DataVolumeReductor based on reductor traitlet"

    subclasses = Factory.child_subclasses(DataVolumeReductor)
    subclass_names = [c.__name__ for c in subclasses]

    reductor = CaselessStrEnum(subclass_names, 'NeighbourPeakIntegrator',
                               help='Data volume reduction scheme to use for '
                                    'the conversion to dl0.').tag(config=True)

    # Product classes traits

    def get_factory_name(self):
        return self.name

    def get_product_name(self):
        return self.reductor



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: UTF-8 -*-
"""
Image timing-based shower image parametrization.
"""

from collections import namedtuple
import numpy as np
from astropy.units import Quantity

__all__ = [
    'TimingParameters',
    'timing_parameters'
]

TimingParameters = namedtuple(
    "TimingParameters",
    "gradient, intercept"
)


class TimingParameterizationError(RuntimeError):
    pass


def rotate_translate(pixel_pos_x, pixel_pos_y, phi):
    """
    Function to perform rotation and translation of pixel lists

    Parameters
    ----------
    pixel_pos_x: ndarray
        Array of pixel x positions
    pixel_pos_y: ndarray
        Array of pixel x positions
    phi: float
        Rotation angle of pixels
        
    Returns
    -------
        ndarray,ndarray: Transformed pixel x and y coordinates

    """

    pixel_pos_rot_x = pixel_pos_x * np.cos(phi) - pixel_pos_y * np.sin(phi)
    pixel_pos_rot_y = pixel_pos_x * np.sin(phi) + pixel_pos_y * np.cos(phi)
    return pixel_pos_rot_x, pixel_pos_rot_y


def timing_parameters(pix_x, pix_y, image, peak_time, rotation_angle):
    """
    Function to extract timing parameters from a cleaned image
    
    Parameters
    ----------
    pix_x : array_like
        Pixel x-coordinate
    pix_y : array_like
        Pixel y-coordinate
    image : array_like
        Pixel values corresponding
    peak_time : array_like
        Pixel times corresponding
    rotation_angle: float
        Rotation angle fo the image major axis

    Returns
    -------
    timing_parameters: TimingParameters
    """

    unit = Quantity(pix_x).unit
    pix_x = Quantity(np.asanyarray(pix_x, dtype=np.float64)).value
    pix_y = Quantity(np.asanyarray(pix_y, dtype=np.float64)).value
    image = np.asanyarray(image, dtype=np.float64)
    peak_time = np.asanyarray(peak_time, dtype=np.float64)

    assert pix_x.shape == image.shape
    assert pix_y.shape == image.shape
    assert peak_time.shape == image.shape

    # Rotate pixels by our image axis
    pix_x_rot, pix_y_rot = rotate_translate(pix_x, pix_y, rotation_angle)
    gradient, intercept = np.polyfit(pix_y_rot, peak_time,deg=1, w=np.sqrt(image))

    return TimingParameters(gradient=gradient*(peak_time.unit/unit),
                            intercept=intercept*unit)



class StagerRep():
    """
    class representing steps status.
    Parameters
    ----------
    name  : str
    next_steps : list(str)
    running : bool
    nb_job_done : int
    """
    STAGER = 1
    PRODUCER = 2
    CONSUMER = 3

    def __init__(self,name,next_steps=None,running=0,
                 nb_job_done=0, queue_length = 0, 
                 nb_process = 1, step_type=STAGER):
        self.type = step_type
        self.name = name
        self.next_steps = next_steps or []
        self.running = running
        self.nb_job_done = nb_job_done
        self.queue_length = queue_length
        self.nb_process = nb_process

    def __repr__(self):
        """  called by the repr() built-in function and by string conversions
        (reverse quotes) to compute the "official" string representation of
        an object.  """
        return (self.name + ' running: '+
            str(self.running)+ '-> nb_job_done: '+
            str(self.nb_job_done) + '-> next_steps:' +
            str(self.next_steps)+ '-> queue_length:' +
            str(self.queue_length)+ '-> nb_process:' +
            str(self.nb_process))

    def get_statistics(self):
        """
        return
        ======
        str containing step name (without its process extension) and the number
        of jobs it did.
        """
        return (self.name.split('$$process')[0] + ' number of jobs done: '+ str(self.nb_job_done))



from astropy.coordinates import Angle
from astropy import units as u

import numpy as np
from numpy import cos, sin, arctan2 as atan2, arccos as acos

__all__ = ['rotate_around_axis', 'rotation_matrix_2d', 'length', 'normalise',
           'angle', 'set_phi_theta', 'set_phi_theta_r']


def rotation_matrix_2d(angle):
    """construct a 2D rotation matrix as a numpy NDArray that rotates a
    vector clockwise. Angle should be any object that can be converted
    into an `astropy.coordinates.Angle`
    """
    psi = Angle(angle).rad
    return np.array([[cos(psi), -sin(psi)],
                     [sin(psi),  cos(psi)]])


def rotate_around_axis(vec, axis, angle):
    """ rotates a vector around an axis by an angle
        creates a rotation matrix and multiplies 
        the initial vector with it

    Parameters
    ----------
    vec  : length-3 numpy array
            3D vector to be rotated
    axis : length-3 numpy array
            axis around which the rotation is performed
    angle : astropy angle quantity or float
            angle (in rad if float) by which vec is rotated around axis

    Returns
    -------
    rotated numpy array
    """

    axis = normalise(axis)
    a = cos(angle / 2.0)
    b, c, d = -axis * sin(angle / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    rot_matrix = np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                           [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                           [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])
    return vec.dot(rot_matrix)


def length(vec):
    """ returns the length/norm of a numpy array
        as the square root of the inner product with itself
    """
    return vec.dot(vec)**.5


def normalise(vec):
    """ Sets the length of the vector to 1
        without changing its direction

    Parameters
    ----------
    vec : numpy array

    Returns
    -------
    numpy array with the same direction but length of 1
    """
    try:
        return vec / length(vec)
    except ZeroDivisionError:
        return vec


def angle(v1, v2):
    """ computes the angle between two vectors
        assuming carthesian coordinates

    Parameters
    ----------
    vec1 : numpy array
    vec2 : numpy array

    Returns
    -------
    the angle between vec1 and vec2 as a dimensioned astropy quantity
    """
    return acos(np.clip(v1.dot(v2) / (length(v1) * length(v2)), -1.0, 1.0))


def set_phi_theta_r(phi, theta, r=1):
    """ sets a 3D vector according to the given angles

    Parameters
    ----------
    phi : astropy.Quantity
    theta : astropy.Quantity
    r : (optional)
        the length of the vector
        can have a unit, doesn't have to

    Returns
    -------
    numpy array with the given direction and length
    """
    return np.array([sin(theta) * cos(phi),
                     sin(theta) * sin(phi),
                     cos(theta)]) * r
""" simple alias for set_phi_theta_r with default (unitless) r argument """
set_phi_theta = lambda x, y: set_phi_theta_r(x, y)


def get_phi_theta(vec):
    """ returns a tupel of the phi and theta angles of the given vector
    """
    try:
        return (atan2(vec[1], vec[0]), acos(np.clip(vec[2] / length(vec), -1, 1))) * u.rad
    except ValueError:
        return (0, 0)



from ctapipe.reco.hillas_intersection import HillasIntersection
import astropy.units as u
from numpy.testing import assert_allclose
import numpy as np

def test_intersect():
    """
    Simple test to check the intersection of lines. Try to intersect positions at (0,1) and (1,0)
    with angles perpendicular and test they cross at (0,0)
    """
    hill = HillasIntersection()
    x1 = 0
    y1 = 1
    theta1 = 90*u.deg

    x2 = 1
    y2 = 0
    theta2 = 0*u.deg

    sx,sy = hill.intersect_lines(x1,y1,theta1,x2,y2,theta2)

    assert_allclose(sx, 0, atol=1e-6)
    assert_allclose(sy, 0, atol=1e-6)

def test_parallel():
    """
    Simple test to check the intersection of lines. Try to intersect positions at (0,0) and (0,1)
    with angles parallel and check the behaviour
    """
    hill = HillasIntersection()
    x1 = 0
    y1 = 0
    theta1 = 0*u.deg

    x2 = 1
    y2 = 0
    theta2 = 0*u.deg

    sx,sy = hill.intersect_lines(x1,y1,theta1,x2,y2,theta2)
    assert_allclose(sx, np.nan, atol=1e-6)
    assert_allclose(sy, np.nan, atol=1e-6)

test_intersect()
test_parallel()


#!/usr/bin/env python3
"""

"""
import math

import numpy as np
from astropy import units as u
from iminuit import Minuit

from ctapipe.coordinates import (HorizonFrame,
                                 NominalFrame,
                                 TiltedGroundFrame,
                                 GroundFrame,
                                 project_to_ground)
from ctapipe.image import poisson_likelihood_gaussian
from ctapipe.io.containers import (ReconstructedShowerContainer,
                                   ReconstructedEnergyContainer)
from ctapipe.reco.reco_algorithms import Reconstructor
from ctapipe.utils import TableInterpolator
from ctapipe.instrument import get_atmosphere_profile_functions
from ctapipe.reco.shower_max import ShowerMaxEstimator

from scipy.optimize import minimize, least_squares
from scipy.stats import norm

__all__ = ['ImPACTReconstructor', 'energy_prior', 'xmax_prior']


def guess_shower_depth(energy):
    """
    Simple estimation of depth of shower max based on the expected gamma-ray elongation 
    rate.
    
    Parameters
    ----------
    energy: float
        Energy of the shower in TeV
        
    Returns
    -------
    float: Expected depth of shower maximum
    """
    x_max_exp = 300 * (u.g*u.cm**-2) + \
                93 * (u.g*u.cm**-2) * np.log10(energy.to(u.TeV).value)

    return x_max_exp

def energy_prior(energy, index=-1):

    return -2 * np.log(np.power(energy, index))


def xmax_prior(energy, xmax, width=30):

    x_max_exp = guess_shower_depth(energy)
    diff = xmax.value - x_max_exp

    return -2 * np.log(norm.pdf(diff/width))


class ImPACTReconstructor(Reconstructor):

    """This class is an implementation if the impact_reco Monte Carlo
    Template based image fitting method from [parsons14]_.  This method uses a
    comparision of the predicted image from a library of image
    templates to perform a maximum likelihood fit for the shower axis,
    energy and height of maximum.

    Because this application is computationally intensive the usual
    advice to use astropy units for all quantities is ignored (as
    these slow down some computations), instead units within the class
    are fixed:

    - Angular units in radians
    - Distance units in metres
    - Energy units in TeV

    References
    ----------
    .. [parsons14] Parsons & Hinton, Astroparticle Physics 56 (2014), pp. 26-34

    """

    def __init__(self, root_dir=".", minimiser="minuit", prior=""):

        # First we create a dictionary of image template interpolators
        # for each telescope type
        self.root_dir = root_dir
        self.prediction = dict()

        self.file_names = {"GATE":"GCT_xm_full.fits", "LSTCam":"LST_xm_full.fits",
                           "NectarCam":"MST_xm_full.fits", "FlashCam":"MST_xm_full.fits"}

        # We also need a conversion function from height above ground to depth of maximum
        # To do this we need the conversion table from CORSIKA
        self.thickness_profile, self.altitude_profile = \
            get_atmosphere_profile_functions('paranal')

        # For likelihood calculation we need the with of the pedestal distribution for each pixel
        # currently this is not availible from the calibration, so for now lets hard code it in a dict
        self.ped_table = {"LSTCam": 1.3, "NectarCam": 2.0, "FlashCam": 2.3,"GATE": 1.3}
        self.spe = 0.5 # Also hard code single p.e. distribution width

        # Also we need to scale the impact_reco templates a bit, this will be fixed later
        self.scale = {"LSTCam": 1.3, "NectarCam": 1.1, "FlashCam": 1.4, "GATE": 1.0}

        self.last_image = dict()
        self.last_point = dict()

        # Next we need the position, area and amplitude from each pixel in the event
        # making this a class member makes passing them around much easier

        self.pixel_x = 0
        self.pixel_y = 0
        self.pixel_area = 0
        self.image = 0
        self.type = ("LST")
        # We also need telescope positions
        self.tel_pos_x = 0
        self.tel_pos_y = 0
        # And the peak of the images
        self.peak_x = 0
        self.peak_y = 0
        self.peak_amp = 0
        self.hillas = 0

        self.ped = dict()

        self.array_direction = 0
        self.minimiser_name = minimiser

        self.array_return = False
        self.priors = prior

    def initialise_templates(self, tel_type):
        """Check if templates for a given telescope type has been initialised
        and if not do it and add to the dictionary

        Parameters
        ----------
        tel_type: dictionary
            Dictionary of telescope types in event

        Returns
        -------
        boolean: Confirm initialisation

        """
        for t in tel_type:
            if tel_type[t] in self.prediction.keys():
                continue

            self.prediction[tel_type[t]] = \
                TableInterpolator(self.root_dir + "/" +
                                  self.file_names[tel_type[t]])

        return True

    def get_brightest_mean(self, num_pix=3):
        """This is a simple function to find the peak position of each image
        in an event which will be used later in the Xmax
        calculation. Peak is found by taking the average position of
        the n hottest pixels in the image.

        Parameters
        ----------
        num_pix: int
            Number of pixels the average position from

        Returns
        -------
            None

        """
        peak_x = np.zeros(
            [len(self.pixel_x)])  # Create blank arrays for peaks
                                  # rather than a dict (faster)
        peak_y = np.zeros(peak_x.shape)
        peak_amp = np.zeros(peak_x.shape)

        # Loop over all tels to take weighted average of pixel
        # positions This loop could maybe be replaced by an array
        # operation by a numpy wizard

        tel_num = 0
        for tel in self.hillas:
            top_index = self.image[tel].argsort()[-1*num_pix:][::-1]

            weight = self.hillas[tel].size
            weighted_x = self.hillas[tel].cen_x.to(u.rad).value * weight
            weighted_y = self.hillas[tel].cen_y.to(u.rad).value * weight

            ppx = np.sum(weighted_x) / np.sum(weight)
            ppy = np.sum(weighted_y) / np.sum(weight)

            peak_x[tel_num] = ppx  # Fill up array
            peak_y[tel_num] = ppy
            peak_amp[tel_num] = np.sum(weight)
            tel_num += 1

        self.peak_x = peak_x  # * unit # Add to class member
        self.peak_y = peak_y  # * unit
        self.peak_amp = peak_amp

    # This function would be useful elsewhere so probably be implemented in a
    # more general form
    def get_shower_max(self, source_x, source_y, core_x, core_y, zen):
        """Function to calculate the depth of shower maximum geometrically
        under the assumption that the shower maximum lies at the
        brightest point of the camera image.

        Parameters
        ----------
        source_x: float
            Event source position in nominal frame
        source_y: float
            Event source position in nominal frame
        core_x: float
            Event core position in telescope tilted frame
        core_y: float
            Event core position in telescope tilted frame
        zen: float
            Zenith angle of event

        Returns
        -------
        float: Depth of maximum of air shower

        """

        # Calculate displacement of image centroid from source position (in
        # rad)
        disp = np.sqrt(np.power(self.peak_x - source_x, 2) +
                       np.power(self.peak_y - source_y, 2))
        # Calculate impact parameter of the shower
        impact = np.sqrt(np.power(np.array(list(self.tel_pos_x.values()))
                                  - core_x, 2) +
                         np.power(np.array(list(self.tel_pos_y.values()))
                                  - core_y, 2))

        # Distance above telescope is ratio of these two (small angle)

        height = impact / disp
        weight = np.power(self.peak_amp,0.)  # weight average by amplitude
        hm = height*u.m
        hm[hm>99*u.km] = 99*u.km

        # Take weighted mean of esimates
        mean_height = np.sum(height * weight) / np.sum(weight)
        # This value is height above telescope in the tilted system,
        # we should convert to height above ground
        mean_height *= np.cos(zen)

        # Add on the height of the detector above sea level
        mean_height += 2100

        if mean_height > 100000 or np.isnan(mean_height):
            mean_height = 100000

        mean_height *= u.m
        # Lookup this height in the depth tables, the convert Hmax to Xmax
        x_max = self.thickness_profile(mean_height.to(u.km))#self.shower_max.interpolate(mean_height.to(u.km))

        # Convert to slant depth
        x_max /= np.cos(zen)
        return x_max

    @staticmethod
    def rotate_translate(pixel_pos_x, pixel_pos_y, x_trans, y_trans, phi):
        """
        Function to perform rotation and translation of pixel lists

        Parameters
        ----------
        pixel_pos_x: ndarray
            Array of pixel x positions
        pixel_pos_y: ndarray
            Array of pixel x positions
        x_trans: float
            Translation of position in x coordinates
        y_trans: float
            Translation of position in y coordinates
        phi: float
            Rotation angle of pixels
        
        Returns
        -------
            ndarray,ndarray: Transformed pixel x and y coordinates

        """

        pixel_pos_trans_x = (pixel_pos_x - x_trans) * \
            np.cos(phi) - (pixel_pos_y - y_trans) * np.sin(phi)
        pixel_pos_trans_y = (pixel_pos_x - x_trans) * \
            np.sin(phi) + (pixel_pos_y - y_trans) * np.cos(phi)
        return pixel_pos_trans_x, pixel_pos_trans_y

    def image_prediction(self, type, zenith, azimuth, energy, impact, x_max, pix_x, pix_y):
        """Creates predicted image for the specified pixels, interpolated
        from the template library.

        Parameters
        ----------
        type: string
            Telescope type specifier
        zenith: float
            Zenith angle of observations
        azimuth: float
            Azimuth angle of observations
        energy: float
            Event energy (TeV)
        impact: float
            Impact diance of shower (metres)
        x_max: float
            Depth of shower maximum (num bins from expectation)
        pix_x: ndarray
            X coordinate of pixels
        pix_y: ndarray
            Y coordinate of pixels

        Returns
        -------
        ndarray: predicted amplitude for all pixels

        """

        return self.prediction[type].interpolate([energy, impact,
                                                  x_max], pix_x, pix_y)

    def get_prediction(self, tel_id, shower_reco, energy_reco):

        horizon_seed = HorizonFrame(az=shower_reco.az, alt=shower_reco.alt)
        nominal_seed = horizon_seed.transform_to(
            NominalFrame(array_direction=horizon_seed))
        source_x = nominal_seed.x.to(u.rad).value
        source_y = nominal_seed.y.to(u.rad).value

        ground = GroundFrame(x=shower_reco.core_x, y=shower_reco.core_y, z=0*u.m)
        tilted = ground.transform_to(
            TiltedGroundFrame(pointing_direction=self.array_direction))
        tilt_x = tilted.x.to(u.m).value
        tilt_y = tilted.y.to(u.m).value

        zenith = 90*u.deg - self.array_direction.alt
        azimuth = self.array_direction.az

        x_max = shower_reco.h_max / np.cos(zenith)

        # Calculate expected Xmax given this energy
        x_max_exp = guess_shower_depth(energy_reco.energy)

        # Convert to binning of Xmax, addition of 100 can probably be removed
        x_max_bin = x_max - x_max_exp

        # Check for range
        if x_max_bin > 250 * (u.g*u.cm**-2):
            x_max_bin = 250 * (u.g*u.cm**-2)
        if x_max_bin < -250 * (u.g*u.cm**-2):
            x_max_bin = -250 * (u.g*u.cm**-2)

        x_max_bin = x_max_bin.value

        impact = np.sqrt(pow(self.tel_pos_x[tel_id] - tilt_x, 2) +
                         pow(self.tel_pos_y[tel_id] - tilt_y, 2))

        phi = np.arctan2( (self.tel_pos_y[tel_id] - tilt_y),
                          (self.tel_pos_x[tel_id] - tilt_x))

        pix_x_rot, pix_y_rot = self.rotate_translate(self.pixel_x[tel_id]
                                                     * -1,
                                                     self.pixel_y[tel_id],
                                                     source_x,
                                                     source_y, phi)

        prediction = self.image_prediction(self.type[tel_id],
                                           (90 * u.deg) - shower_reco.alt,
                                           shower_reco.az,
                                           energy_reco.energy.value,
                                           impact, x_max_bin,
                                           pix_x_rot * (180 / math.pi),
                                           pix_y_rot * (180 / math.pi))

        prediction *= self.scale[self.type[tel_id]]
        #prediction *= self.pixel_area[tel_id]

        prediction[prediction < 0] = 0
        prediction[np.isnan(prediction)] = 0

        return prediction

    def get_likelihood(self, source_x, source_y, core_x, core_y,
                       energy, x_max_scale):
        """Get the likelihood that the image predicted at the given test
        position matches the camera image.

        Parameters
        ----------
        source_x: float
            Source position of shower in the nominal system (in deg)
        source_y: float
            Source position of shower in the nominal system (in deg)
        core_x: float
            Core position of shower in tilted telescope system (in m)
        core_y: float
            Core position of shower in tilted telescope system (in m)
        energy: float
            Shower energy (in TeV)
        x_max_scale: float
            Scaling factor applied to geometrically calculated Xmax

        Returns
        -------
        float: Likelihood the model represents the camera image at this position

        """

        # First we add units back onto everything.  Currently not
        # handled very well, maybe in future we could just put
        # everything in the correct units when loading in the class
        # and ignore them from then on

        zenith = 90*u.deg - self.array_direction.alt
        azimuth = self.array_direction.az

        # Geometrically calculate the depth of maximum given this test position
        x_max = self.get_shower_max(source_x, source_y,
                                    core_x, core_y,
                                    zenith.to(u.rad).value) * x_max_scale
        # Calculate expected Xmax given this energy
        x_max_exp = guess_shower_depth(energy*u.TeV)

        # Convert to binning of Xmax, addition of 100 can probably be removed
        x_max_bin = x_max - x_max_exp

        # Check for range
        if x_max_bin > 250 * (u.g*u.cm**-2):
            x_max_bin = 250 * (u.g*u.cm**-2)
        if x_max_bin < -250 * (u.g*u.cm**-2):
            x_max_bin = -250 * (u.g*u.cm**-2)

        x_max_bin = x_max_bin.value

        array_like = None

        for tel_count in self.image:  # Loop over all telescopes
            # Calculate impact distance for all telescopes
            impact = np.sqrt(pow(self.tel_pos_x[tel_count] - core_x, 2)
                             + pow(self.tel_pos_y[tel_count] - core_y, 2))
            # And the expected rotation angle
            phi = np.arctan2((self.tel_pos_y[tel_count] - core_y),
                             (self.tel_pos_x[tel_count] - core_x))  # - (math.pi/2.)

            # Rotate and translate all pixels such that they match the
            # template orientation
            pix_x_rot, pix_y_rot = self.rotate_translate(
                self.pixel_x[tel_count] * -1,
                self.pixel_y[tel_count],
                source_x, source_y, phi
            )

            # Then get the predicted image, convert pixel positions to deg
            prediction = self.image_prediction(
                self.type[tel_count],
                zenith, azimuth, energy, impact, x_max_bin,
                pix_x_rot * (180 / math.pi),
                pix_y_rot * (180 / math.pi)
            )
            prediction[np.isnan(prediction)] = 0
            prediction[prediction < 1e-6] = 1e-6

            # Scale templates to match simulations
            prediction *= self.scale[self.type[tel_count]]
            #prediction *= self.pixel_area[tel_count]

            # Get likelihood that the prediction matched the camera image
            like = poisson_likelihood_gaussian(self.image[tel_count],
                                               prediction,
                                               self.spe,
                                               self.ped[tel_count])
            if np.any(prediction == np.inf):
                print("inf found at ", self.type[tel_count], zenith,
                      azimuth, energy, impact, x_max_bin)
            like[np.isnan(like)] = 1e9
            if array_like is None:
                array_like = like
            else:
                array_like = np.append(array_like,like)

        prior_pen = 0
        # Add prior penalities if we have them
        array_like += 1e-8
        if "energy" in self.priors:
            prior_pen += energy_prior(energy, index=-2)
        if "xmax" in self.priors:
            prior_pen += xmax_prior(energy, x_max)

        array_like += prior_pen/float(len(array_like))
        if self.array_return:
            return array_like
        return np.sum(array_like)


    def get_likelihood_min(self, x):
        """Wrapper class around likelihood function for use with scipy
        minimisers

        Parameters
        ----------
        x: ndarray
            Array of minimisation parameters

        Returns
        -------
        float: Likelihood value of test position

        """
        return self.get_likelihood(x[0], x[1], x[2], x[3], x[4], x[5])

    def set_event_properties(self, image, pixel_x, pixel_y,
                             pixel_area, type_tel, tel_x, tel_y,
                             array_direction, hillas):
        """The setter class is used to set the event properties within this
        class before minimisation can take place. This simply copies a
        bunch of useful properties to class members, so that we can
        use them later without passing all this information around.

        Parameters
        ----------
        image: dictionary
            Amplitude of pixels in camera images
        pixel_x: dictionary
            X position of pixels in nominal system
        pixel_y: dictionary
            Y position of pixels in nominal system
        pixel_area: dictionary
            Area of pixel in each telescope type
        type_tel: dictionary
            Type of telescope
        tel_x: dictionary
            X position of telescope
        tel_y: dictionary
            Y position of telescope

        Returns
        -------
        None

        """
        # First store these parameters in the class so we can use them
        # in minimisation For most values this is simply copying
        self.image = image

        self.pixel_x = dict()
        self.pixel_y = dict()

        self.tel_pos_x = dict()
        self.tel_pos_y = dict()
        self.pixel_area = dict()
        self.ped = dict()

        # So here we must loop over the telescopes
        for x in tel_x:
            self.pixel_x[x] = pixel_x[x].to(u.rad).value
            self.pixel_y[x] = pixel_y[x].to(u.rad).value

            self.tel_pos_x[x] = tel_x[x].value
            self.tel_pos_y[x] = tel_y[x].value
            self.pixel_area[x] = pixel_area[x].to(u.deg*u.deg).value
            # Here look up pedestal value
            self.ped[x] = self.ped_table[type_tel[x]]

        self.hillas = hillas

        self.get_brightest_mean(num_pix=3)
        self.type = type_tel
        self.initialise_templates(type_tel)

        self.array_direction = array_direction
        self.last_image = 0
        self.last_point = 0

    def predict(self, shower_seed, energy_seed):
        """
        
        Parameters
        ----------
        shower_seed: ReconstructedShowerContainer
            Seed shower geometry to be used in the fit
        energy_seed: ReconstructedEnergyContainer
            Seed energy to be used in fit

        Returns
        -------
        ReconstructedShowerContainer, ReconstructedEnergyContainer:
        Reconstructed ImPACT shower geometry and energy
        
        """
        horizon_seed = HorizonFrame(az=shower_seed.az, alt=shower_seed.alt)
        nominal_seed = horizon_seed.transform_to(NominalFrame(array_direction=self.array_direction))
        print(nominal_seed)
        print(horizon_seed)
        print(self.array_direction)

        source_x = nominal_seed.x[0].to(u.rad).value
        source_y = nominal_seed.y[0].to(u.rad).value

        ground = GroundFrame(x=shower_seed.core_x,
                             y=shower_seed.core_y, z=0 * u.m)
        tilted = ground.transform_to(
            TiltedGroundFrame(pointing_direction=self.array_direction)
        )
        tilt_x = tilted.x.to(u.m).value
        tilt_y = tilted.y.to(u.m).value

        lower_en_limit = energy_seed.energy * 0.5
        en_seed =  energy_seed.energy
        if lower_en_limit < 0.04 * u.TeV:
            lower_en_limit = 0.04 * u.TeV
            en_seed = 0.041 * u.TeV

        seed = (source_x, source_y, tilt_x,
                tilt_y, en_seed.value, 0.8)
        step = (0.001, 0.001, 10, 10, en_seed.value*0.1, 0.1)
        limits = ((source_x-0.01, source_x+0.01),
                  (source_y-0.01, source_y+0.01),
                  (tilt_x-100, tilt_x+100),
                  (tilt_y-100, tilt_y+100),
                  (lower_en_limit.value, en_seed.value*2),
                  (0.5,2))

        fit_params, errors = self.minimise(params=seed, step=step, limits=limits,
                                             minimiser_name=self.minimiser_name)

        # container class for reconstructed showers '''
        shower_result = ReconstructedShowerContainer()

        nominal = NominalFrame(x=fit_params[0] * u.rad,
                               y=fit_params[1] * u.rad,
                               array_direction=self.array_direction)
        horizon = nominal.transform_to(HorizonFrame())

        shower_result.alt, shower_result.az = horizon.alt, horizon.az
        tilted = TiltedGroundFrame(x=fit_params[2] * u.m,
                                   y=fit_params[3] * u.m,
                                   pointing_direction=self.array_direction)
        ground = project_to_ground(tilted)

        shower_result.core_x = ground.x
        shower_result.core_y = ground.y

        shower_result.is_valid = True

        shower_result.alt_uncert = np.nan
        shower_result.az_uncert = np.nan
        shower_result.core_uncert = np.nan

        zenith = 90*u.deg - self.array_direction.alt
        shower_result.h_max = fit_params[5] * \
                              self.get_shower_max(fit_params[0],
                                                  fit_params[1],
                                                  fit_params[2],
                                                  fit_params[3],
                                                  zenith.to(u.rad).value)

        shower_result.h_max_uncert = errors[5] * shower_result.h_max

        shower_result.goodness_of_fit = np.nan
        shower_result.tel_ids = list(self.image.keys())

        energy_result = ReconstructedEnergyContainer()
        energy_result.energy = fit_params[4] * u.TeV
        energy_result.energy_uncert = errors[4] * u.TeV
        energy_result.is_valid = True
        energy_result.tel_ids = list(self.image.keys())
        # Return interesting stuff

        return shower_result, energy_result

    def minimise(self, params, step, limits, minimiser_name="minuit"):
        """
        
        Parameters
        ----------
        params
        step
        limits
        minimiser_name

        Returns
        -------

        """
        if minimiser_name == "minuit":

            min = Minuit(self.get_likelihood,
                         print_level=1,
                         source_x=params[0],
                         error_source_x=step[0],
                         limit_source_x=limits[0],
                         source_y=params[1],
                         error_source_y=step[1],
                         limit_source_y=limits[1],
                         core_x=params[2],
                         error_core_x=step[2],
                         limit_core_x=limits[2],
                         core_y=params[3],
                         error_core_y=step[3],
                         limit_core_y=limits[3],
                         energy=params[4],
                         error_energy=step[4],
                         limit_energy=limits[4],
                         x_max_scale=params[5], error_x_max_scale=step[5],
                         limit_x_max_scale=limits[5],
                         fix_x_max_scale=False,
                         errordef=1)

            min.tol *= 1000
            min.set_strategy(0)

            # Perform minimisation
            migrad = min.migrad()
            fit_params = min.values
            errors = min.errors

            return (fit_params["source_x"], fit_params["source_y"], fit_params["core_x"],
                    fit_params["core_y"], fit_params["energy"],fit_params[
                        "x_max_scale"]),\
                   (errors["source_x"], errors["source_y"], errors["core_x"],
                    errors["core_x"], errors["energy"],errors["x_max_scale"])

        elif minimiser_name == "lm" or minimiser_name == "trf" or\
                        minimiser_name == "dogleg":
            self.array_return = True
            limits = np.array(limits)

            min = least_squares(self.get_likelihood_min,params,
                                method=minimiser_name,
                                x_scale=step,
                                xtol=1e-10,
                                ftol=1e-10
                                )
            return min.x, (0,0,0,0,0,0)

        else:
            min = minimize(self.get_likelihood_min,params,
                           method=minimiser_name,
                           bounds=limits
                           )
            print(min.x)
            return min.x, (0,0,0,0,0,0)

    def draw_nominal_surface(self, shower_seed, energy_seed, bins=30,
                             nominal_range=2.5*u.deg):
        """
        Simple reconstruction for evaluating the likelihood in a grid across the 
        nominal system, fixing all values but the source position of the gamma rays. 
        Useful for checking the reconstruction performance of the algorithm
        
        Parameters
        ----------
        shower_seed: ReconstructedShowerContainer
            Best fit ImPACT shower geometry 
        energy_seed: ReconstructedEnergyContainer
            Best fit ImPACT energy
        bins: int
            Number of bins in surface evaluation
        nominal_range: Quantity
            Range over which to create likelihood surface

        Returns
        -------
        ndarray, ndarray, ndarray: 
        Bin centres in X and Y coordinates and the values of the likelihood at each 
        position
        """
        horizon_seed = HorizonFrame(az=shower_seed.az, alt=shower_seed.alt)
        nominal_seed = horizon_seed.transform_to(
            NominalFrame(array_direction=self.array_direction))

        source_x = nominal_seed.x[0].to(u.rad)
        source_y = nominal_seed.y[0].to(u.rad)

        ground = GroundFrame(x=shower_seed.core_x,
                             y=shower_seed.core_y, z=0 * u.m)
        tilted = ground.transform_to(
            TiltedGroundFrame(pointing_direction=self.array_direction)
        )
        tilt_x = tilted.x.to(u.m)
        tilt_y = tilted.y.to(u.m)

        x_dir = np.linspace(source_x - nominal_range, source_x + nominal_range, num=bins)
        y_dir = np.linspace(source_y - nominal_range, source_y + nominal_range, num=bins)
        w = np.zeros([bins, bins])
        zenith = 90*u.deg - self.array_direction.alt

        for xb in range(bins):
            for yb in range(bins):
                x_max_scale = shower_seed.h_max / \
                              self.get_shower_max(x_dir[xb].to(u.rad).value,
                                                  y_dir[yb].to(u.rad).value,
                                                  tilt_x.value,
                                                  tilt_y.value,
                                                  zenith.to(u.rad).value)

                w[xb][yb] = self.get_likelihood(x_dir[xb].to(u.rad).value,
                                                y_dir[yb].to(u.rad).value,
                                                tilt_x.value,
                                                tilt_y.value,
                                                energy_seed.energy.value, x_max_scale)

        w = w - np.min(w)

        return x_dir.to(u.deg), y_dir.to(u.deg), w

    def draw_tilted_surface(self, shower_seed, energy_seed,
                            bins=50, core_range=100 * u.m):
        """
        Simple reconstruction for evaluating the likelihood in a grid across the 
        nominal system, fixing all values but the core position of the gamma rays. 
        Useful for checking the reconstruction performance of the algorithm
        
        Parameters
        ----------
        shower_seed: ReconstructedShowerContainer
            Best fit ImPACT shower geometry 
        energy_seed: ReconstructedEnergyContainer
            Best fit ImPACT energy
        bins: int
            Number of bins in surface evaluation
        nominal_range: Quantity
            Range over which to create likelihood surface

        Returns
        -------
        ndarray, ndarray, ndarray: 
        Bin centres in X and Y coordinates and the values of the likelihood at each 
        position
        """
        horizon_seed = HorizonFrame(az=shower_seed.az, alt=shower_seed.alt)
        nominal_seed = horizon_seed.transform_to(
            NominalFrame(array_direction=self.array_direction))

        source_x = nominal_seed.x[0].to(u.rad).value
        source_y = nominal_seed.y[0].to(u.rad).value

        ground = GroundFrame(x=shower_seed.core_x,
                             y=shower_seed.core_y, z=0 * u.m)
        tilted = ground.transform_to(
            TiltedGroundFrame(pointing_direction=self.array_direction)
        )
        tilt_x = tilted.x.to(u.m)
        tilt_y = tilted.y.to(u.m)

        x_ground_list = np.linspace(tilt_x - core_range, tilt_x + core_range, num=bins)
        y_ground_list = np.linspace(tilt_y - core_range, tilt_y + core_range, num=bins)
        w = np.zeros([bins, bins])
        zenith = 90*u.deg - self.array_direction.alt

        for xb in range(bins):
            for yb in range(bins):
                x_max_scale = shower_seed.h_max / \
                              self.get_shower_max(source_x,
                                                  source_y,
                                                  x_ground_list[xb].value,
                                                  y_ground_list[yb].value,
                                                  zenith.to(u.rad).value)

                w[xb][yb] =  self.get_likelihood(source_x,
                                                 source_y,
                                                 x_ground_list[xb].value,
                                                 y_ground_list[yb].value,
                                                 energy_seed.energy.value, x_max_scale)
        return x_ground_list, y_ground_list, w



from ctapipe.core import Component


class RingFitter(Component):
    """This is the base class from which all ring fitters should inherit from"""

    def fit(self, x, y, weight, times=None):
        """overwrite this method with your favourite ring fitting algorithm

        Parameters
        ----------
        x: array
           vector of pixel x-coordinates as astropy quantities
        y: array
           vector of pixel y-coordinates as astropy quantities
        weight: array
           vector of pixel weights
        times: array
           optional vector of pixel DAQ times as astropy quantities

        Returns
        -------
        tuple of centre_x, centre_y, radius  as astropy quantities

        """
        pass



from ctapipe.reco.regressor_classifier_base import *

from sklearn.ensemble import RandomForestClassifier


def test_reshuffle_event_list():

    feature_list = [
            {"FlashCam": [[1, 10], [2, 20]],
             "ASTRICam": [[30, 3], [40, 4]]},
            {"FlashCam": [[1.5, 15]],
             "ASTRICam": [[35, 3.5], [25, 2.5], [30, 3]]}
    ]
    target_list = ["1", "2"]

    cam_id_list = ["FlashCam", "ASTRICam"]
    my_base = RegressorClassifierBase(model=RandomForestClassifier,
                                      cam_id_list=cam_id_list, unit=1)

    feature_flattened, target_flattened = my_base.reshuffle_event_list(feature_list,
                                                                       target_list)

    assert feature_flattened == {'FlashCam': [[1, 10], [2, 20], [1.5, 15]],
                                 'ASTRICam': [[30, 3], [40, 4], [35, 3.5], [25, 2.5],
                                              [30, 3]]}

    assert target_flattened == {'FlashCam': ['1', '1', '2'],
                                'ASTRICam': ['1', '1', '2', '2', '2']}



from ctapipe.core import Component
from time import sleep

class Odd(Component):
    """Odd` class represents a Stage for pipeline.
    It returns received value to Inverse stage except when
    received value is a multiple of 5. In this case it returns
    received value to Add stage
    """
    def init(self):
        self.log.debug("--- Odd init ---")
        return True

    def run(self, _input):
        sleep(.5)
        self.log.debug("ODD receive {}".format(_input))
        if _input % 5 == 0:
            return (_input, 'Add')
        else:
            return (_input, 'Inverse')

    def finish(self):
        self.log.debug("--- Odd finish ---")
        pass



"""
Calibrator for the R1 -> DL0 data level transition.

This module handles the calibration from the R1 data level to DL0. This
transition exists as a conveniance in the pipepline and can be used to test
data volume reduction methods inside the pipeline. By default, no data volume
reduction is applied, and the DL0 samples are identical to the R1. However,
if a reductor from `ctapipe.image.reductors` is passed to the
`CameraDL0Reducer`, then the reduction will be applied.
"""
from ctapipe.core import Component

__all__ = ['CameraDL0Reducer']


class CameraDL0Reducer(Component):
    """
    Parent class for the dl0 data volume reducers. Fills the dl0 container.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool or None
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    reductor : ctapipe.calib.camera.reductors.Reductor
        The reductor to use to reduce the waveforms in the event.
        By default no data volume reduction is applied, and the dl0 samples
        will equal the r1 samples.
    kwargs
    """

    name = 'CameraDL0Reducer'

    def __init__(self, config, tool, reductor=None, **kwargs):
        super().__init__(config=config, parent=tool, **kwargs)
        if reductor is None:
            self.log.info("Applying no data volume reduction in the "
                          "conversion from R1 to DL0")
        self._reductor = reductor
        self._r1_empty_warn = False

    def check_r1_exists(self, event, telid):
        """
        Check that r1 data exists. If it does not, then do not change dl0.

        This ensures that if the containers were filled from a file containing
        r1 data, it is not overwritten by non-existant data.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        telid : int
            The telescope id.

        Returns
        -------
        bool
            True if r1.tel[telid].pe_samples is not None, else false.
        """
        r1 = event.r1.tel[telid].pe_samples
        if r1 is not None:
            return True
        else:
            if not self._r1_empty_warn:
                self.log.warning("Encountered an event with no R1 data. "
                                 "DL0 is unchanged in this circumstance.")
                self._r1_empty_warn = True
            return False

    def reduce(self, event):
        """
        Perform the conversion from raw R1 data to dl0 data
        (PE Samples -> Reduced PE Samples), and fill the dl0 container.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        """
        tels = event.r1.tels_with_data
        for telid in tels:
            r1 = event.r1.tel[telid].pe_samples
            if self.check_r1_exists(event, telid):
                if self._reductor is None:
                    event.dl0.tel[telid].pe_samples = r1
                else:
                    reduction = self._reductor.reduce_waveforms(r1)
                    event.dl0.tel[telid].pe_samples = reduction
